"""
EPyT-C Test Part 3
This file is provided to ensure that all functions can be executed correctly using Net3.inp 
and MSRT-2: Bacterial regrowth module.
"""
from epyt import epanet
d = epanet('Net3.inp') # load Net3.inp
from epytc.bacterial_regrowth import module
import numpy as np

class TestEpytcModule2():
    """
    Class to test the functions used in bacterial_regrowth.py
    """
    def test_details(self):
        """
        Display modulde details
        """
        assert module.details() != ''
        print('Function 1 test - Success')
        
    def test_network(self, d):
        """
        Getting basic details of the network
        """
        assert module.network(d)[0] == 97 # count of nodes
        assert module.network(d)[1] == 119 # count of links
        assert module.network(d)[2] == 2 # count of reservoirs
        assert module.network(d)[3] == 3 # count of tanks
        assert module.network(d)[4] == 2 # count of pumps
        assert module.network(d)[5] == 0 # count of valves
        assert module.network(d)[6][2] == '20' # name of the third node
        assert module.network(d)[7][5] == '103' # name of the sixth link
        assert module.network(d)[8] == [93, 94] # indices of the two reservoirs
        assert module.network(d)[9] == [95, 96, 97] # indices of the three tanks
        assert module.network(d)[10] == [118, 119] # indices of the two pumps
        assert module.network(d)[11] == [] # indices of the valves (no valves present)
        assert module.network(d)[12] == 'GPM' # flow unit
        assert module.network(d)[13][5, 0] == 10 # start node of the sixth link
        assert module.network(d)[13][5, 1] == 11 # end node of the sixth link
        assert module.network(d)[14][9, 0] == 1 # true connectivity between 1st and 10th nodes
        assert module.network(d)[14][10, 0] == 0 # false connectivity between 1st and 11th nodes
        assert module.network(d)[15][5] == 411.48 # length of the sixth pipe in SI unit (m)
        assert module.network(d)[16][5] == 406.4 # diameter of the sixth pipe in SI unit (mm)
        assert module.network(d)[17][5] == 10 # start node of the sixth link
        assert module.network(d)[18][5] == 11 # end node of the sixth link
        assert module.network(d)[19] == 0 # count of nodes omitted from the analysis
        assert module.network(d)[20] == [] # indices of omitted nodes
        assert module.network(d)[21] == 0 # count of links omitted from the analysis
        assert module.network(d)[22] == [] # indices of omitted links
        print('Function 2 test - Success')
        
    def test_species(self):
        """
        Defining the species information of the MSRT module selected
        """
        assert module.species()[0] == 5 # count of water quality parameters
        assert module.species()[1] == 5 # count of bulk water quality parameters
        assert module.species()[2] == 0 # count of wall water quality parameters
        assert module.species()[3] == 19 # count of model variables
        print('Function 3 test - Success')
        
    def test_zero_order_reaction(self, water_quality_step):
        """
        Defining zero-order reaction
        """
        water_quality_step = 300 # water quality time step (s)
        assert module.zero_order_reaction(water_quality_step) == 300
        print('Function 4 test - Success')
        
    def test_first_order_reaction(self, reaction_rate_constant, concentration_value, water_quality_step):
        """
        Defining first-order reaction
        Using dummy data
        """
        reaction_rate_constant = -1e-3 # (1/s)
        concentration_value = 1 # (mg/L)
        water_quality_step = 300 # water quality time step (s)
        assert round(module.first_order_reaction(reaction_rate_constant, concentration_value, water_quality_step), 1) == -0.2
        print('Function 5 test - Success')
        
    def test_Reynolds_number(self, pipe_velocity, pipe_diameter, kinematic_viscosity):
        """
        Defining Reynolds number
        Using dummy data
        """
        pipe_velocity = 0.2 # (m/s)
        pipe_diameter = 200 # (mm)
        kinematic_viscosity = 9.31e-7 # (sq.m/s)
        assert round(module.Reynolds_number(pipe_velocity, pipe_diameter, kinematic_viscosity)) == 42965
        print('Function 6 test - Success')
        
    def test_Schmidt_number(self, kinematic_viscosity, molecular_diffusivity):
        """
        Defining Schmidt number
        Using dummy data
        """
        kinematic_viscosity = 9.31e-7 # (sq.m/s)
        molecular_diffusivity = 12.5e-10 # (sq.m/s)
        assert round(module.Schmidt_number(kinematic_viscosity, molecular_diffusivity)) == 745
        print('Function 7 test - Success')
        
    def test_Sherwood_number(self, Reynolds_number, Schmidt_number, pipe_diameter, pipe_length):
        """
        Defining Sherwood number
        Using dummy data
        """
        Reynolds_number = 42965
        Schmidt_number = 745
        pipe_diameter = 200 # (mm)
        pipe_length = 1000 # (m)
        assert round(module.Sherwood_number(Reynolds_number, Schmidt_number, pipe_diameter, pipe_length)) == 33
        print('Function 8 test - Success')
        
    def test_mass_transfer_coefficient(self, Sherwood_number, molecular_diffusivity, pipe_diameter):
        """
        Defining Mass transfer coefficient
        Using dummy data
        """
        Sherwood_number = 33
        molecular_diffusivity = 12.5e-10 # (sq.m/s)
        pipe_diameter = 200 # (mm)
        assert round(module.mass_transfer_coefficient(Sherwood_number, molecular_diffusivity, pipe_diameter), 7) == 2e-7
        print('Function 9 test - Success')
                             
    def test_hydraulic_mean_radius(self, pipe_diameter):
        """
        Defining Defining Hydrauic mean radius
        Using dummy data
        """
        pipe_diameter = 200 # (mm)
        assert module.hydraulic_mean_radius(pipe_diameter) == 0.05
        print('Function 10 test - Success')
        
    def test_variables(self, maximum_iterations_count, variables_count):
        """
        Defining variables of the MSRT module
        """
        maximum_iterations_count = 200 # times for which the water qality simulation will be iteratively performed
        variables_count = 19 # count of selected MSRT model variables
        assert (module.variables(maximum_iterations_count, variables_count)[:, 0] > 0).all()
        assert np.logical_and(5.2e-6 <= module.variables(maximum_iterations_count, variables_count)[:, 1], \
                       module.variables(maximum_iterations_count, variables_count)[:, 1] <= 3.4e-3).all()
        assert np.logical_and(1.04e-7 <= module.variables(maximum_iterations_count, variables_count)[:, 2], \
                       module.variables(maximum_iterations_count, variables_count)[:, 2] <= 1.43e-5).all()
        assert np.logical_and(0.15 <= module.variables(maximum_iterations_count, variables_count)[:, 3], \
                       module.variables(maximum_iterations_count, variables_count)[:, 3] <= 2.50).all()
        assert np.logical_and(-12.6 <= module.variables(maximum_iterations_count, variables_count)[:, 4], \
                       module.variables(maximum_iterations_count, variables_count)[:, 4] <= -7.0).all()
        assert np.logical_and(-3 <= module.variables(maximum_iterations_count, variables_count)[:, 5], \
                       module.variables(maximum_iterations_count, variables_count)[:, 5] <= 0.19).all()  
        assert (module.variables(maximum_iterations_count, variables_count)[:, 6] <= 5).all()
        assert (module.variables(maximum_iterations_count, variables_count)[:, 7] <= 6e-3).all()
        assert (module.variables(maximum_iterations_count, variables_count)[:, 8] <= 0.15).all()
        assert np.logical_and(-5 <= module.variables(maximum_iterations_count, variables_count)[:, 10], \
                       module.variables(maximum_iterations_count, variables_count)[:, 10] <= 0.5).all()
        assert (module.variables(maximum_iterations_count, variables_count)[:, 11] <= 2.2e-6).all()
        assert (module.variables(maximum_iterations_count, variables_count)[:, 12] <= 46).all()
        assert (module.variables(maximum_iterations_count, variables_count)[:, 13] == 12.5e-10).all()
        assert np.logical_and(7.8e-10 <= module.variables(maximum_iterations_count, variables_count)[:, 14], \
                       module.variables(maximum_iterations_count, variables_count)[:, 14] <= 11.5e-10).all()
        assert np.logical_and(6.7e-10 <= module.variables(maximum_iterations_count, variables_count)[:, 15], \
                       module.variables(maximum_iterations_count, variables_count)[:, 15] <= 12.1e-10).all()
        assert np.logical_and(8e-10 <= module.variables(maximum_iterations_count, variables_count)[:, 16], \
                       module.variables(maximum_iterations_count, variables_count)[:, 16] <= 34e-10).all()
        assert (module.variables(maximum_iterations_count, variables_count)[:, 17] == 1e-13).all()
        assert (module.variables(maximum_iterations_count, variables_count)[:, 18] == 9.31e-7).all()
        print('Function 11 test - Success')
        
    def test_pipe_reaction(self, water_quality_step, pipe_number, grid_number, pipe_velocity, pipe_diameter,\
                           pipe_length, pipe_segment_width, variables_matrix, pipe_concentration_matrix):
        """
        Defining pipe reactions for the MSRT model selected
        Using dummy data
        """
        water_quality_step = 300 # water quality time step (s)
        pipe_number = 1 # 2nd pipe is considered
        grid_number = 2 # 3rd grid
        pipe_velocity = 0.5 # (m/s)
        pipe_diameter = 200 # (mm)
        pipe_length = 1000 # (m)
        pipe_segment_width = 16.667 # (m)
        variables_matrix = module.variables(200, 19)[2] # Variable values corresponding to 3rd iteration
        pipe_concentration_matrix = np.ones((5, 5, 119)) # Assuming the maxiumum segments is 5
        assert (module.pipe_reaction(water_quality_step, pipe_number, grid_number, pipe_velocity, pipe_diameter, \
                             pipe_length, pipe_segment_width, variables_matrix, pipe_concentration_matrix)[0] <= 0)
        assert (module.pipe_reaction(water_quality_step, pipe_number, grid_number, pipe_velocity, pipe_diameter, \
                                 pipe_length, pipe_segment_width, variables_matrix, pipe_concentration_matrix)[1] <= 0)
        assert (module.pipe_reaction(water_quality_step, pipe_number, grid_number, pipe_velocity, pipe_diameter, \
                                 pipe_length, pipe_segment_width, variables_matrix, pipe_concentration_matrix)[2] != 'Nan')
        assert (module.pipe_reaction(water_quality_step, pipe_number, grid_number, pipe_velocity, pipe_diameter, \
                                 pipe_length, pipe_segment_width, variables_matrix, pipe_concentration_matrix)[3] != 'Nan')
        assert (module.pipe_reaction(water_quality_step, pipe_number, grid_number, pipe_velocity, pipe_diameter, \
                                 pipe_length, pipe_segment_width, variables_matrix, pipe_concentration_matrix)[4] != 'Nan')
        print('Function 12 test - Success')
        
    def test_tank_reaction(self, water_quality_step, water_quality_step_number, tank_number, tank_volume_previous_step,\
                           tank_volume_current, variables_matrix, initial_node_concentration_matrix, node_concentration_matrix):
       """
       Defining tank reactions for the MSRT model selected
       Using dummy data
       """
       water_quality_step = 300 # water quality time step (s)
       water_quality_step_number = 10 # 10th step of simulation
       tank_number = 96 # 97th node or 3rd tank
       tank_volume_previous_step = 19358 # tank volume in cu.m for the previous step
       tank_volume_current = 20172 # current tank volume in cu.m
       variables_matrix = module.variables(200, 19)[2] # Variable values corresponding to 3rd iteration
       initial_node_concentration_matrix = np.zeros((97, 5))
       node_concentration_matrix = np.ones((5, 2881, 97))
       assert (module.tank_reaction(water_quality_step, water_quality_step_number, tank_number, tank_volume_previous_step,\
                              tank_volume_current, variables_matrix, initial_node_concentration_matrix, \
                                  node_concentration_matrix)[0] <= 0)
       assert (module.tank_reaction(water_quality_step, water_quality_step_number, tank_number, tank_volume_previous_step,\
                             tank_volume_current, variables_matrix, initial_node_concentration_matrix, \
                                 node_concentration_matrix)[1] <= 0)
       assert (module.tank_reaction(water_quality_step, water_quality_step_number, tank_number, tank_volume_previous_step,\
                           tank_volume_current, variables_matrix, initial_node_concentration_matrix, \
                               node_concentration_matrix)[2] != 'Nan')
       assert (module.tank_reaction(water_quality_step, water_quality_step_number, tank_number, tank_volume_previous_step,\
                           tank_volume_current, variables_matrix, initial_node_concentration_matrix, \
                               node_concentration_matrix)[3] != 'Nan')
       assert (module.tank_reaction(water_quality_step, water_quality_step_number, tank_number, tank_volume_previous_step,\
                           tank_volume_current, variables_matrix, initial_node_concentration_matrix, \
                               node_concentration_matrix)[4] != 'Nan')
       print('Function 13 test - Success')
    
    def test_reservoir_quality(self, d, maximum_iterations_count, percentage_variation_random_pattern1, source_quality_matrix, \
                               pattern_input_command1):
        """
        Defining source quality at the reservoir(s)
        Using dummy data
        """
        maximum_iterations_count = 200 # times for which the water qality simulation will be iteratively performed
        percentage_variation_random_pattern1 = 0, 0.2 # 0 and 20% variations
        source_quality_matrix = [[1, 1, 0.5, 10, 0],[1, 1, 0.5, 10, 0]] # quality at the two sources of Net3 (boundary condition)
        pattern_input_command1 = 'none', 'rand' 
        
        val = module.reservoir_quality(d, maximum_iterations_count, percentage_variation_random_pattern1[0], source_quality_matrix, \
                                pattern_input_command1[0])
        assert (val[0] == source_quality_matrix[0]).all()
        assert (val[1] == source_quality_matrix[1]).all()
        
        val = module.reservoir_quality(d, maximum_iterations_count, percentage_variation_random_pattern1[1], source_quality_matrix, \
                                pattern_input_command1[1])
        assert np.logical_and(0.8 <= val[0][:, 0], val[0][:, 0] <= 1.2).all()    
        assert np.logical_and(0.8 <= val[0][:, 1], val[0][:, 1] <= 1.2).all()
        assert np.logical_and(0.4 <= val[0][:, 2], val[0][:, 2] <= 0.6).all()
        assert np.logical_and(8 <= val[0][:, 3], val[0][:, 3] <= 12).all()
        assert (val[0][:, 4] == 0).all()
        assert np.logical_and(0.8 <= val[1][:, 0], val[1][:, 0] <= 1.2).all()    
        assert np.logical_and(0.8 <= val[1][:, 1], val[1][:, 1] <= 1.2).all()
        assert np.logical_and(0.4 <= val[1][:, 2], val[1][:, 2] <= 0.6).all()
        assert np.logical_and(8 <= val[1][:, 3], val[1][:, 3] <= 12).all()
        assert (val[1][:, 4] == 0).all()
        print('Function 14 test - Success')
        
    def test_reservoir_pattern(self, d, base_time_days, percentage_variation_random_pattern2, reservoir_injection_start_step, \
                               reservoir_injection_end_step, reservoir_injection_input, pattern_input_command2):
        """
        Defining source quality pattern for the reservoir(s)
        Using dummy data
        """
        base_time_days = 1 # base time period (day)
        percentage_variation_random_pattern2 = 0, 0.2 # 0 and 20% variations
        reservoir_injection_start_step = [[0],[12]] # injection for first an second source nodes start at 0th and 12th h, respectively
        reservoir_injection_end_step = [[12], [24]] # injection for first an second source nodes end at 12th and 24th h, respectively
        reservoir_injection_input = [[1], [1]] # value to multiply the source quality
        pattern_input_command2 = 'none', 'rand', 'specific'
        
        val = module.reservoir_pattern(d, base_time_days, percentage_variation_random_pattern2[0], reservoir_injection_start_step, \
                                   reservoir_injection_end_step, reservoir_injection_input, pattern_input_command2[0])
        assert (val[0] == [1, 1, 1, 1, 1]).all()
        assert (val[1] == [1, 1, 1, 1, 1]).all()
        
        val = module.reservoir_pattern(d, base_time_days, percentage_variation_random_pattern2[1], reservoir_injection_start_step, \
                                   reservoir_injection_end_step, reservoir_injection_input, pattern_input_command2[1])
        assert np.logical_and(0.8 <= val, val <= 1.2).all()
        
        val = module.reservoir_pattern(d, base_time_days, percentage_variation_random_pattern2[1], reservoir_injection_start_step, \
                                   reservoir_injection_end_step, reservoir_injection_input, pattern_input_command2[2])
        assert (val[0][reservoir_injection_start_step[0][0]: reservoir_injection_end_step[0][0]] == reservoir_injection_input[0][0]).all()
        assert (val[0][reservoir_injection_start_step[1][0]: reservoir_injection_end_step[1][0]] == 0).all()
        assert (val[1][reservoir_injection_start_step[0][0]: reservoir_injection_end_step[0][0]] == 0).all()
        assert (val[1][reservoir_injection_start_step[1][0]: reservoir_injection_end_step[1][0]] == reservoir_injection_input[0][0]).all()
        print('Function 15 test - Success')
        
    def test_injection_quality(self, maximum_iterations_count, percentage_variation_random_pattern1, injection_node_index_matrix, \
                               injection_node_quality_matrix, pattern_input_command1):
        """
        Defining quality at the injection node(s)
        Using dummy data
        """
        maximum_iterations_count = 200 # times for which the water qality simulation will be iteratively performed
        percentage_variation_random_pattern1 = 0, 0.2 # 0 and 20% variations
        injection_node_index_matrix = [21] # node with index 21 is specified as injection node
        injection_node_quality_matrix = [[2, 0, 0, 0, 0]] # # quality at the injection node of Net3 (boundary condition)
        pattern_input_command1 = 'none', 'rand' 
        val = module.injection_quality(maximum_iterations_count, percentage_variation_random_pattern1[0], injection_node_index_matrix, \
                                   injection_node_quality_matrix, pattern_input_command1[0])
        assert (val == injection_node_quality_matrix[0]).all()
        
        val = module.injection_quality(maximum_iterations_count, percentage_variation_random_pattern1[0], injection_node_index_matrix, \
                                   injection_node_quality_matrix, pattern_input_command1[0])
        assert np.logical_and(1.6 <= val[0][:, 0], val[0][:, 0] <= 2.4).all()
        assert (val[0][:,1: 5] == 0).all()
        print('Function 16 test - Success')
    
    def test_injection_pattern(self, d, base_time_days, percentage_variation_random_pattern2, injection_node_index_matrix, injection_node_injection_start_step, \
                               injection_node_injection_end_step, injection_node_injection_input, pattern_input_command2):
        """
        Defining injection pattern for the injection node(s)
        Using dummy data
        """
        base_time_days = 1 # base time period (day)
        percentage_variation_random_pattern2 = 0, 0.2 # 0 and 20% variations
        injection_node_index_matrix = [21] # node with index 21 is specified as injection node
        injection_node_injection_start_step = [[3]] # injection for injection node start at 3rd h
        injection_node_injection_end_step = [[19]] # injection for injection node end at 19th h
        injection_node_injection_input = [[1]] # value to multiply the injection node quality
        pattern_input_command2 = 'none', 'rand', 'specific'
        
        val = module.injection_pattern(d, base_time_days, percentage_variation_random_pattern2[0], injection_node_index_matrix, injection_node_injection_start_step, \
                                   injection_node_injection_end_step, injection_node_injection_input, pattern_input_command2[0])
        assert (val == [1, 1, 1, 1, 1]).all()
        
        val = module.injection_pattern(d, base_time_days, percentage_variation_random_pattern2[1], injection_node_index_matrix, injection_node_injection_start_step, \
                                   injection_node_injection_end_step, injection_node_injection_input, pattern_input_command2[1])
        assert np.logical_and(0.8 <= val, val <= 1.2).all()
        
        val = module.injection_pattern(d, base_time_days, percentage_variation_random_pattern2[1], injection_node_index_matrix, injection_node_injection_start_step, \
                                   injection_node_injection_end_step, injection_node_injection_input, pattern_input_command2[2])
        assert (val[injection_node_injection_start_step[0][0]: injection_node_injection_end_step[0][0]] == injection_node_injection_input[0][0]).all()
        assert (val[: injection_node_injection_start_step[0][0], 0] == 0).all()
        assert (val[injection_node_injection_end_step[0][0] :, 0] == 0).all()
        print('Function 17 test - Success')

test = TestEpytcModule2()
test.test_details()
test.test_network(d)
test.test_species()
test.test_zero_order_reaction(300)
test.test_first_order_reaction(1e-3, 1, 300)
test.test_Reynolds_number(0.2, 200, 9.31e-7)
test.test_Schmidt_number(9.31e-7, 12.5e-10)
test.test_Sherwood_number(42965, 745, 200, 1000)
test.test_mass_transfer_coefficient(33, 12.5e-10, 200)
test.test_hydraulic_mean_radius(200)
test.test_variables(200, 19)
test.test_pipe_reaction(300, 1, 2, 0.5, 200, 100, 16.667, module.variables(200, 19)[2], np.ones((5, 5, 119)))
test.test_tank_reaction(300, 10, 96, 19358, 20172, module.variables(200, 19)[2], np.zeros((97, 5)), np.ones((5, 2881, 97)))
test.test_reservoir_quality(d, 200, (0, 0.2), [[1, 1, 0.5, 10, 0],[1, 1, 0.5, 10, 0]], ('none', 'rand'))
test.test_reservoir_pattern(d, 1, (0, 0.2), [[0],[12]], [[12], [24]], [[1], [1]], ('none', 'rand', 'specific'))
test.test_injection_quality(200, (0, 0.2), [21], [[2, 0, 0, 0, 0]], ('none', 'rand'))
test.test_injection_pattern(d, 1, (0, 0.2), [21], [[3]], [[19]], [[1]], ('none', 'rand', 'specific'))

print('\nAll tests completed!\n')


"""Arsenite oxidation and arsenate attachment/detachment module
Reactive species (bulk) - Aqueous arsenite (mg/L), Aqueous arsenate (mg/L), Aqueous arsenic (mg/L), Residual chlorine (mg/L), and Total organic carbon (mg/L)
Reactive species (wall) - Adsorbed arsenate (mg/sq.m)
"""

import math
import random
import numpy as np


class module:
    def details():
        """Displaying the information about the MSRT model selected for water quality analysis

        :return: Details of module
        :rtype: String
        """
        print("Arsenite oxidation and arsenate attachemnt/detachment module loaded.")
        print("\nReactive species (bulk):")
        print(
            "Aqueous arsenite (mg/L)\nAqueous arsenate (mg/L)\nAqueous arsenic (mg/L)\
              \nResidual chlorine (mg/L)\nTotal organic carbon (mg/L)"
        )
        print("\nReactive species (wall):")
        print("Adsorbed arsenate (mg/sq.m)")

    def network(d):
        """Getting basic details of the network

        :param d: EPANET model
        :type d: EPANET object
        :return: Network details
        :rtype: List
        """
        network_info = [
            d.getNodeCount(),
            d.getLinkCount(),
            d.getNodeReservoirCount(),
            d.getNodeTankCount(),
            d.getLinkPumpCount(),
            d.getLinkValveCount(),
            d.getNodeNameID(),
            d.getLinkNameID(),
            d.getNodeReservoirIndex(),
            d.getNodeTankIndex(),
        ]
        index_pumps = []
        for x in range(
            d.getLinkCount() - (d.getLinkPumpCount() + d.getLinkValveCount()),
            len(d.getLinkNameID()) - d.getLinkValveCount(),
        ):
            index_pumps.append(x + 1)
        network_info.append(index_pumps)
        index_valves = []
        for x in range(d.getLinkCount() - d.getLinkValveCount(), len(d.getLinkNameID())):
            index_valves.append(x + 1)
        network_info.append(index_valves)
        network_info.extend(
            [
                d.getFlowUnits(),
                d.getNodesConnectingLinksIndex(),
                d.getConnectivityMatrix(),
            ]
        )
        # Conversion of GPM units to SI units
        if d.getFlowUnits() == "GPM":
            network_info.append(0.3048 * d.getLinkLength())
            network_info.append(25.4 * d.getLinkDiameter())
        else:
            network_info.extend([d.getLinkLength(), d.getLinkDiameter()])
        start_node_matrix = []
        end_node_matrix = []
        for x in range(d.getLinkCount()):
            var1 = d.getNodesConnectingLinksIndex()[x][0]
            var2 = d.getNodesConnectingLinksIndex()[x][1]
            start_node_matrix.append(var1)
            end_node_matrix.append(var2)
        network_info.extend([start_node_matrix, end_node_matrix])
        # Number of nodes omitted for analysis, if any
        number_omitted_nodes = 0
        # Index of omitted nodes
        index_omitted_nodes = []
        # Number of links omitted for analysis, if any
        number_omitted_links = 0
        # Index of omitted nodes
        index_omitted_links = []
        network_info.extend(
            [
                number_omitted_nodes,
                index_omitted_nodes,
                number_omitted_links,
                index_omitted_links,
            ]
        )
        return network_info
    
    def species():
        """Defining the species information of the MSRT module selected

        :return: Species information of the MSRT module
        :rtype: String
        """
        msrt_info = []
        number_water_quality_parameters = 6
        msrt_info.append(number_water_quality_parameters)
        number_bulk_water_quality_parameters = 5
        msrt_info.append(number_bulk_water_quality_parameters)
        number_wall_water_quality_parameters = 1
        msrt_info.append(number_wall_water_quality_parameters)
        number_model_variables = 14
        msrt_info.append(number_model_variables)
        msrt_info = [
            number_water_quality_parameters,
            number_bulk_water_quality_parameters,
            number_wall_water_quality_parameters,
            number_model_variables,
        ]
        return msrt_info

    def zero_order_reaction(num1):
        """Defining zero-order reaction

        :param num1: Water quality simulation time step in seconds
        :type num1: Integer
        :return: Solution of zero-order ordinary differential equation
        :rtype: Float
        """
        delta_zero_order = num1
        return delta_zero_order

    def first_order_reaction(num1, num2, num3):
        """Defining first-order reaction

        :param num1: Reaction rate constant
        :type num1: Float
        :param num2: Concentration value
        :type num2: Float
        :param num3: Water quality simulation time step in seconds
        :type num3: Integer
        :return: Solution of first-order ordinary differential equation
        :rtype: Float
        """
        m1 = num1 * num1
        m2 = num1 * (num2 + (num3 / 4) * m1)
        m3 = num1 * (num2 + (num3 / 4) * m2)
        m4 = num1 * (num2 + (num3 / 2) * m3)
        delta_first_order = (num3 / 6) * (m1 + 2 * m2 + 2 * m3 + m4)
        return delta_first_order

    def Reynolds_number(num1, num2, num3):
        """Defining Reynolds number

        :param num1: Pipe flow velocity in metres per second
        :type num1: Float
        :param num2: Pipe diameter in millimetres
        :type num2: Float
        :param num3: Kinematic viscosity of water in square metres per second
        :type num3: Float
        :return: Reynolds number 
        :rtype: Float
        """
        num4 = num2 * 1e-3
        reynolds_num = (num1 * num4) / num3
        return reynolds_num

    def Schmidt_number(num1, num2):
        """Defining Schmidt number

        :param num1: Kinematic viscosity of water in square metres per second
        :type num1: Float
        :param num2: Molecular diffusivity of a bulk phase species in square metres per second
        :type num2: Float
        :return: Schmidt number 
        :rtype: Float
        """
        schmidt_num = num1 / num2
        return schmidt_num

    def Sherwood_number(num1, num2, num3, num4):
        """Defining Sherwood number

        :param num1: Reynolds number
        :type num1: Float
        :param num2: Schmidt number
        :type num2: Float
        :return: Schmidt number
        :param num3: Pipe diameter in millimetres
        :type num3: Float
        :param num4: Pipe length in metres
        :type num4: Float
        :return: Sherwood number
        :rtype: Float
        """
        num5 = num3 * 1e-3
        if num1 < 2300:
            sherwood_num = 0.023 * (num1**0.83) * (num2**0.33)
        else:
            sherwood_num = 3.65 + (
                (0.0668 * (num5 / num4) * num1 * num2)
                / (1 + 0.04 * ((num5 / num4) * num1 * num2) ** (2 / 3))
            )
        return sherwood_num

    def mass_transfer_coefficient_cl(num1, num2, num3):
        """Defining mass-transfer coefficient for chlorine

        :param num1: Sherwood number
        :type num1: Float
        :param num2: Molecular diffusivity of chlorine in square metres per second
        :type num2: Float
        :return: Schmidt number
        :param num3: Pipe diameter in millimetres
        :type num3: Float
        :return: Mass-transfer coefficient for chlorine
        :rtype: Float
        """
        num4 = num3 * 1e-3
        kf_val_cl = num1 * (num2 / num4)
        return kf_val_cl

    def mass_transfer_coefficient_ars(num1, num2, num3, num4):
        """Defining mass-transfer coefficient for arsenic

        :param num1: Molecular diffusivity of arsenic in square metres per second
        :type num1: Float
        :param num2: Pipe diameter in millimetres
        :type num2: Float
        :return: Schmidt number
        :param num3: Pipe flow velocity is metres per second
        :type num3: Float
        :param num4: Kinematic viscosity of water in square metres per second
        :type num4: Float
        :return: Mass-transfer coefficient for arsenic
        :rtype: Float
        """
        num5 = num2 * 1e-3
        if num5 == 0:
            kf_val_ars = 1
        else:
            kf_val_ars = 0.026 * (num1 / num5) * ((num5 * num3 / num4) ** 0.8) * ((num4 / num1) ** 0.333)
        return kf_val_ars

    def hydraulic_mean_radius(num1):
        """Defining hydraulic mean radius

        :param num1: Pipe diameter in millimetres
        :type num1: Float
        :return: Hydraulic mean radius
        :rtype: Float
        """
        num2 = num1 * 1e-3
        rh_value = num2 / 4
        return rh_value

    def area_per_unit_vol(num1):
        """Defining area of croos-section

        :param num1: Pipe diameter in millimetres
        :type num1: Float
        :return: Area of cross-section
        :rtype: Float
        """
        num2 = num1 * 1e-3
        area_val = 4 / (1e3 * num2)
        return area_val
    
    def variables(num1, num2):
        """Defining variables of the MSRT module

        :param num1: Number of iterations
        :type num1: Integer
        :param num2: Number of variables corresponding to the MSRT module selected
        :type num2: Integer
        :return: Matrix of variable values
        :rtype: Array
        """
        variable_mat = np.zeros((num1, num2))
        # Temperature (degree Celsius)
        temperature_mean = 25
        temperature_var = 0.5
        # Rate constant (chlorine - TOC reaction) (L/mg-C/s)
        kbNC_lower = 2.19e4
        kbNC_upper = 3.81e4
        kbNC_mean = 3e4
        # Rate constant (chlorine wall reaction) (m/s)
        kwC_lower = 1.04e-7
        kwC_upper = 1.43e-5
        kwC_mean = 1.22e-6
        # Reaction yield constant (chlorine - TOC reaction) (mg-C/ mg-Cl)
        YN_upper = 0.15
        YN_lower = 2.50
        YN_mean = 0.61
        # Adsorbance rate constant (for arsenic) (L/mg/s)
        k_ads = 6.67e-5
        # Equilibrium constant (for arsenic) (mg/L)
        k_eq = 0.054
        # Maximum adsorbance density (for As(V)) (mg/sq.m)
        S_max = 500
        # Rate constant (chlorine - As(III) reaction) (L/mg/s)
        kbCAs = 3.64
        # Reaction yield constant (chlorine - As(III) reaction) (mg-Cl/ mg-As(III))
        YC = 0.96
        # Molecular diffusivity of chlorine (sq.m/s)
        Dm_chlorine = 12.5e-10
        # Molecular diffusivity of TOC (sq.m/s)
        Dm_toc_lower = 7.8e-10
        Dm_toc_upper = 11.5e-10
        Dm_toc_mean = 9.5e-10
        # Molecular diffusivity of arsenic (sq.m/s)
        Dm_arsenic = 7.5e-10
        # Molecular diffusivity of THMs (sq.m/s)
        dens_water = 997
        # Kinematic viscosity of water (sq.m/s)
        nu_water = 9.31e-7
        if num1 == 1:
            variable_mat[num1 - 1][0] = temperature_mean
            variable_mat[num1 - 1][1] = kbNC_mean * math.exp(-6050 / (temperature_mean + 273))
            variable_mat[num1 - 1][2] = kwC_mean
            variable_mat[num1 - 1][3] = YN_mean
            variable_mat[num1 - 1][4] = k_ads
            variable_mat[num1 - 1][5] = k_eq
            variable_mat[num1 - 1][6] = S_max
            variable_mat[num1 - 1][7] = kbCAs
            variable_mat[num1 - 1][8] = YC
            variable_mat[num1 - 1][9] = Dm_chlorine
            variable_mat[num1 - 1][10] = Dm_toc_mean
            variable_mat[num1 - 1][11] = Dm_arsenic
            variable_mat[num1 - 1][12] = dens_water
            variable_mat[num1 - 1][13] = nu_water
        else:
            for x in range(num1):
                variable_mat[x][0] = (1 - temperature_var) * temperature_mean + (
                    2 * temperature_var * temperature_mean
                ) * random.uniform(0, 1)
                variable_mat[x][1] = random.uniform(kbNC_lower, kbNC_upper) * math.exp(
                    -6050 / ((variable_mat[x][0]) + 273)
                )
                variable_mat[x][2] = random.uniform(kwC_lower, kwC_upper)
                variable_mat[x][3] = random.uniform(YN_lower, YN_upper)
                variable_mat[x][4] = k_ads
                variable_mat[x][5] = k_eq
                variable_mat[x][6] = S_max
                variable_mat[x][7] = kbCAs
                variable_mat[x][8] = YC
                variable_mat[x][9] = Dm_chlorine
                variable_mat[x][10] = random.uniform(Dm_toc_lower, Dm_toc_upper)
                variable_mat[x][11] = Dm_arsenic
                variable_mat[x][12] = dens_water
                variable_mat[x][13] = nu_water
        return variable_mat

    def pipe_reaction(num1, num2, num3, num4, num5, num6, num7, arr1, arr2):
        """Defining link reactions for the MSRT model selected

        :param num1: Water quality simulation time step in seconds
        :type num1: Integer
        :param num2: Link index
        :type num2: Integer
        :param num3: Grid index
        :type num3: Integer
        :param num4: Link flow velocity in metres per second
        :type num3: Float
        :param num5: Link diameter in millimetres
        :type num5: Float
        :param num6: Link length in metres
        :type num6: Float
        :param num7: Link segment length in metres
        :type num7: Float
        :param arr1: List of variable values
        :type arr1: List
        :param arr2: Matrix of link concentration values
        :type arr2: Array
        :return: Values corresponding to growth or decay of the concentration of water quality parameters
        :rtype: List
        """
        kbNC = arr1[1]
        kwC = arr1[2]
        YN = arr1[3]
        k_ads = arr1[4]
        k_eq = arr1[5]
        S_max = arr1[6]
        kbCAs = arr1[7]
        YC = arr1[8]
        Dm_chlorine = arr1[9]
        Dm_arsenic = arr1[11]
        nu_water = arr1[13]

        KbNC = kbNC * arr2[4][num3][num2]
        KbCAs = kbCAs * arr2[3][num3][num2]
        Re = module.Reynolds_number(num4, num5, nu_water)
        Sc_chlorine = module.Schmidt_number(nu_water, Dm_chlorine)
        Sh_chlorine = module.Sherwood_number(Re, Sc_chlorine, num5, num6)
        kfC = module.mass_transfer_coefficient_cl(Sh_chlorine, Dm_chlorine, num5)
        rh = module.hydraulic_mean_radius(num5)
        KwC = (kwC * kfC) / ((kwC + kfC) * rh)
        KfAs = module.mass_transfer_coefficient_ars(Dm_arsenic, num2, num4, nu_water)
        Sorb = (arr2[1][num3][num2] * (S_max - arr2[5][num3][num2]) - k_eq * arr2[5][num3][num2]) / (
            (1 / k_ads) + ((1 / KfAs) * (S_max - arr2[5][num3][num2]))
        )
        Av = module.area_per_unit_vol(num5)

        # Reactions within pipe
        delta_chlorine_toc_reac_pipe = module.first_order_reaction(KbNC, arr2[3][num3][num2], num1)
        delta_chlorine_wall_reac_pipe = module.first_order_reaction(KwC, arr2[3][num3][num2], num1)
        delta_toc_chlorine_reac_pipe = YN * delta_chlorine_toc_reac_pipe
        if module.first_order_reaction(KbCAs, arr2[0][num3][num2], num1) > arr2[0][num3][num2]:
            delta_As3_chlorine_reac_pipe = arr2[0][num3][num2]
        else:
            delta_As3_chlorine_reac_pipe = module.first_order_reaction(KbCAs, arr2[0][num3][num2], num1)
        delta_chlorine_As3_reac_pipe = YC * delta_chlorine_toc_reac_pipe
        if module.first_order_reaction(KbCAs, arr2[0][num3][num2], num1) > arr2[0][num3][num2]:
            delta_As5_As3_reac_pipe = arr2[0][num3][num2]
        else:
            delta_As5_As3_reac_pipe = module.first_order_reaction(KbCAs, arr2[0][num3][num2], num1)
        # delta_As5_As3_reac_pipe = 0
        delta_As5_bulk_reac_pipe = Av * Sorb
        delta_As5_wall_reac_pipe = Sorb

        net_delta_chlorine_reac = (
            -delta_chlorine_toc_reac_pipe - delta_chlorine_wall_reac_pipe - delta_chlorine_As3_reac_pipe
        )
        net_delta_toc_reac = -delta_toc_chlorine_reac_pipe
        net_delta_As3_reac = -delta_As3_chlorine_reac_pipe
        net_delta_As5_reac_bulk = delta_As5_As3_reac_pipe - delta_As5_bulk_reac_pipe
        net_delta_As5_reac_wall = delta_As5_wall_reac_pipe
        net_delta_As_reac = net_delta_As3_reac + net_delta_As5_reac_bulk

        delta_mat = [
            net_delta_As3_reac,
            net_delta_As5_reac_bulk,
            net_delta_As_reac,
            net_delta_chlorine_reac,
            net_delta_toc_reac,
            net_delta_As5_reac_wall,
        ]
        return delta_mat

    def tank_reaction(num1, num2, num3, num4, num5, arr1, arr2, arr3):
        """Defining tank reactions for the MSRT model selected

        :param num1: Water quality simulation time step in seconds
        :type num1: Integer
        :param num2: Present water quality step
        :type num2: Integer
        :param num3: Tank index
        :type num3: Integer
        :param num4: Tank volume in previous water quality step
        :type num4: Integer
        :param num5: Tank volume in the present water quality step
        :type num5: Integer
        :param arr1: List of variable values
        :type arr1: List
        :param arr2: Matrix of initial tank concentration values
        :type arr2: Array
        :param arr2: Matrix of tank concentration values
        :type arr2: Array
        :return: Values corresponding to growth or decay of the concentration of water quality parameters
        :rtype: List
        """
        kbNC = arr1[1]
        YN = arr1[3]
        kbCAs = arr1[7]
        YC = arr1[8]

        if num2 == 1:
            tank_As3_conc = arr2[num3][0]
            tank_chlorine_conc = arr2[num3][3]
            tank_toc_conc = arr2[num3][4]
        else:
            tank_As3_conc = arr3[0][num2 - 1][num3]
            tank_chlorine_conc = arr3[3][num2 - 1][num3]
            tank_toc_conc = arr3[4][num2 - 1][num3]

        KbNC = kbNC * tank_toc_conc
        KbCAs = kbCAs * tank_chlorine_conc
        # Reactions within tank
        delta_chlorine_toc_reac_tank = (num4 / num5) * module.first_order_reaction(
            KbNC, tank_chlorine_conc, num1
        )
        delta_toc_chlorine_reac_tank = YN * delta_chlorine_toc_reac_tank
        if (num4 / num5) * module.first_order_reaction(KbCAs, tank_As3_conc, num1) > tank_As3_conc:
            delta_As3_chlorine_reac_tank = tank_As3_conc
        else:
            delta_As3_chlorine_reac_tank = (num4 / num5) * module.first_order_reaction(
                KbCAs, tank_As3_conc, num1
            )
        delta_chlorine_As3_reac_tank = YC * delta_As3_chlorine_reac_tank

        net_delta_chlorine_reac = -delta_chlorine_toc_reac_tank - delta_chlorine_As3_reac_tank
        net_delta_toc_reac = -delta_toc_chlorine_reac_tank
        net_delta_As3_reac = -delta_As3_chlorine_reac_tank
        net_delta_As5_reac = delta_As3_chlorine_reac_tank
        net_delta_As_reac = 0

        delta_mat = [
            net_delta_As3_reac,
            net_delta_As5_reac,
            net_delta_As_reac,
            net_delta_chlorine_reac,
            net_delta_toc_reac,
        ]
        return delta_mat

    def reservoir_quality(d, num1, num2, arr1, str1):
        """Defining source quality values for the reservoir(s)

        :param num1: Number of iterations
        :type num1: Integer
        :param num2: Variability in the random pattern for source quality
        :type num2: Float
        :param arr1: Quality values for the reservoir(s)
        :type arr1: Array
        :param str1: Input command for the random pattern
        :type str1: String
        :return: Values corresponding to source quality at the reservoir(s)
        :rtype: Array
        """
        num_reservoirs = module.network(d)[2]
        num_bulk_parameters = module.species()[1]
        if len(arr1) == num_reservoirs:
            if len(arr1[0]) == num_bulk_parameters:
                print("Reservoir quality updated.")
            else:
                print("Reservoir quality input error.")
                exit()
        reservoir_quality = np.zeros((num_reservoirs, num1, num_bulk_parameters))
        input = str1
        rand_vary = num2
        if input == "none":
            for x in range(num_reservoirs):
                reservoir_quality[x] = arr1[x]
        elif input == "rand":
            if num1 == 1:
                mat_min = arr1
                mat_max = arr1
            else:
                mat_min = np.multiply(arr1, (1 - rand_vary))
                mat_max = np.multiply(arr1, (1 + rand_vary))
            for x in range(num_reservoirs):
                for y in range(num_bulk_parameters):
                    z = 0
                    while z < num1:
                        reservoir_quality[x][z][y] = random.uniform(mat_min[x][y], mat_max[x][y])
                        z += 1
        return reservoir_quality

    def reservoir_pattern(d, num1, num2, arr1, arr2, arr3, str1):
        """Defining source quality pattern for the reservoir(s)

        :param num1: Base time period in day(s)
        :type num1: Float/Integer
        :param num2: Variability in the pattern
        :type num2: Float
        :param arr1: Start time step for the injection in the reservoir(s)
        :type arr1: Array
        :param arr2: End time step for the injection in the reservoir(s)
        :type arr2: Array
        :param arr3: Input value for the injection in the reservoir(s)
        :type arr3: Array
        :param str1: Input command for the pattern
        :type str1: String
        :return: Values corresponding to source quality pattern at the reservoir(s)
        :rtype: Array
        """
        num_reservoirs = module.network(d)[2]
        num_bulk_parameters = module.species()[1]
        h_time = d.getTimeHydraulicStep()
        pattern_steps = int(num1 * 24 * 3600 / h_time)
        pattern_mat = np.zeros((num_reservoirs, pattern_steps, num_bulk_parameters))
        # Input
        input = str1
        rand_vary = num2
        if input == "none":
            pattern_mat = np.add(pattern_mat, 1)
        elif input == "rand":
            for x in range(num_reservoirs):
                for y in range(num_bulk_parameters):
                    z = 0
                    while z < pattern_steps:
                        pattern_mat[x][z][y] = random.uniform(1 - rand_vary, 1 + rand_vary)
                        z += 1
        elif input == "specific":
            start_step_mat = arr1
            end_step_mat = arr2
            val_input = arr3
            if len(start_step_mat) == num_reservoirs and len(end_step_mat) == num_reservoirs:
                if len(start_step_mat[0]) <= pattern_steps and len(end_step_mat[0]) <= pattern_steps:
                    for x in range(num_reservoirs):
                        for y in range(len(start_step_mat[x])):
                            pattern_mat[x][start_step_mat[x][y] : end_step_mat[x][y]] = val_input[x][y]
            else:
                exit()
        return pattern_mat

    def injection_quality(num1, num2, arr1, arr2, str1):
        """Defining source quality values for the injection node(s)

        :param num1: Number of iterations
        :type num1: Integer
        :param num2: Variability in the random pattern for injection node quality
        :type num2: Float
        :param arr1: Index value(s) of injection node(s)
        :type arr1: List
        :param arr2: Quality values for the injection node(s)
        :type arr2: Array
        :param str1: Input command for the random pattern
        :type str1: String
        :return: Values corresponding to source quality at the injection node(s)
        :rtype: Array
        """
        # num1 - number of iterations; arr1 - matrix of injection nodes indices; arr2 - matrix of injection nodes quality
        # str1 - input value for pattern; num2 - percentage variation in the random pattern
        num_injection_nodes = len(arr1)
        num_bulk_parameters = module.species()[1]
        if len(arr2) == num_injection_nodes:
            if len(arr2[0]) == num_bulk_parameters:
                print("Reservoir quality updated.")
            else:
                print("Injection node quality input error.")
                exit()
        injection_quality = np.zeros((num_injection_nodes, num1, num_bulk_parameters))
        print("Injection nodes quality updated.")
        # Input
        input = str1
        rand_vary = num2 
        if input == "none":
            for x in range(num_injection_nodes):
                injection_quality[x] = arr2[x]
        elif input == "rand":
            if num1 == 1:
                mat_min = arr2
                mat_max = arr2
            else:
                mat_min = np.multiply(arr2, (1 - rand_vary))
                mat_max = np.multiply(arr2, (1 + rand_vary))
            for x in range(num_injection_nodes):
                for y in range(num_bulk_parameters):
                    z = 0
                    while z < num1:
                        injection_quality[x][z][y] = random.uniform(mat_min[x][y], mat_max[x][y])
                        z += 1
        return injection_quality

    def injection_pattern(d, num1, num2, arr1, arr2, arr3, arr4, str1):
        """Defining source quality pattern for the injection node(s)

        :param num1: Base time period in day(s)
        :type num1: Float/Integer
        :param num2: Variability in the pattern
        :type num2: Float
        :param arr1: Index value(s) of injection node(s)
        :type arr1: List
        :param arr2: Start time step for the injection in the injection node(s)
        :type arr2: Array
        :param arr3: End time step for the injection in the injection node(s)
        :type arr3: Array
        :param arr4: Input value for the injection in the injection node(s)
        :type arr4: Array
        :param str1: Input command for the pattern
        :type str1: String
        :return: Values corresponding to source quality pattern at the injection node(s)
        :rtype: Array
        """
        num_injection_nodes = len(arr1)
        num_bulk_parameters = module.species()[1]
        h_time = d.getTimeHydraulicStep()
        pattern_steps = int(num1 * 24 * 3600 / h_time)
        inj_pattern_mat = np.zeros((num_injection_nodes, pattern_steps, num_bulk_parameters))
        # Input
        input = str1
        rand_vary = num2
        if input == "none":
            inj_pattern_mat = np.add(inj_pattern_mat, 1)
        elif input == "rand":
            for x in range(num_injection_nodes):
                for y in range(num_bulk_parameters):
                    z = 0
                    while z < pattern_steps:
                        inj_pattern_mat[x][z][y] = random.uniform(1 - rand_vary, 1 + rand_vary)
                        z += 1
        elif input == "specific":
            start_step_mat = arr2
            end_step_mat = arr3
            val_input = arr4
            if len(start_step_mat) == num_injection_nodes and len(end_step_mat) == num_injection_nodes:
                if len(start_step_mat[0]) <= pattern_steps and len(end_step_mat[0]) <= pattern_steps:
                    for x in range(num_injection_nodes):
                        for y in range(len(start_step_mat[x])):
                            inj_pattern_mat[x][start_step_mat[x][y] : end_step_mat[x][y]] = val_input[x][y]
            else:
                exit()
        return inj_pattern_mat



"""
EPyT-C Test Part 1
This file is provided to ensure that all functions can be executed correctly using Net3.inp.
"""
from epytc.functions import fn 
from epyt import epanet
d = epanet('Net3.inp') # load Net3.inp
H = d.getComputedHydraulicTimeSeries() # perform hydraulic simulation using EPANET solver
import numpy as np

class TestEpytcFunctions():
    """
    Class to test the functions used in epytc/functions.py
    """
    def test_network(self, d):
        """
        Getting basic details of the inp file
        """
        assert fn.network(d)[0] == 97 # count of nodes
        assert fn.network(d)[1] == 119 # count of links
        assert fn.network(d)[2] == 2 # count of reservoirs
        assert fn.network(d)[3] == 3 # count of tanks
        assert fn.network(d)[4] == 2 # count of pumps
        assert fn.network(d)[5] == 0 # count of valves
        assert fn.network(d)[6][2] == '20' # name of the third node
        assert fn.network(d)[7][5] == '103' # name of the sixth link
        assert fn.network(d)[8] == [93, 94] # indices of the two reservoirs
        assert fn.network(d)[9] == [95, 96, 97] # indices of the three tanks
        assert fn.network(d)[10] == [118, 119] # indices of the two pumps
        assert fn.network(d)[11] == [] # indices of the valves (no valves present)
        assert fn.network(d)[12] == 'GPM' # flow unit
        assert fn.network(d)[13][5, 0] == 10 # start node of the sixth link
        assert fn.network(d)[13][5, 1] == 11 # end node of the sixth link
        assert fn.network(d)[14][9, 0] == 1 # true connectivity between 1st and 10th nodes
        assert fn.network(d)[14][10, 0] == 0 # false connectivity between 1st and 11th nodes
        assert fn.network(d)[15][5] == 411.48 # length of the sixth pipe in SI unit (m)
        assert fn.network(d)[16][5] == 406.4 # diameter of the sixth pipe in SI unit (mm)
        assert fn.network(d)[17][5] == 10 # start node of the sixth link
        assert fn.network(d)[18][5] == 11 # end node of the sixth link
        assert fn.network(d)[19] == 0 # count of nodes omitted from the analysis
        assert fn.network(d)[20] == [] # indices of omitted nodes
        assert fn.network(d)[21] == 0 # count of links omitted from the analysis
        assert fn.network(d)[22] == [] # indices of omitted links
        print('Function 1 test - Success')
        
    def test_incoming_links(self, node_id, end_nodes_matrix):
        """
        Getting incoming links of a specific node
        """
        end_nodes_matrix = fn.network(d)[18] # end nodes of every Net3 link
        assert fn.incoming_links(node_id, end_nodes_matrix)[0] == [117] # incoming link to the first node
        print('Function 2 test - Success')
        
    def test_outgoing_links(self, node_id, start_nodes_matrix):
        """
        Getting outgoing links of a specific node
        """
        node_id = 1
        start_nodes_matrix = fn.network(d)[17] # start nodes of every Net3 link
        assert fn.incoming_links(node_id, start_nodes_matrix)[0] == [30] # outgoing link from the second node
        print('Function 3 test - Success')
        
    def test_minimum_link_length(self, steps_count, water_quality_step, pipe_velocity_tolerance, pipe_velocity_matrix):
        """
        Getting minimum length (m) to be specified for pumps and valves
        """
        steps_count = 27 # time steps in the hydraulic report
        water_quality_step = 300 # water quality time step (s)
        pipe_velocity_tolerance = 0.001 # tolerable flow velocity considered (m/s)
        pipe_velocity_matrix = H.Velocity # from EPANET
        assert fn.minimum_link_length(steps_count, water_quality_step, pipe_velocity_tolerance, pipe_velocity_matrix) == 0.3
        print('Function 4 test - Success')
        
    def test_minimum_link_diameter(self, links_count, pumps_count, valves_count, pipe_diameter_matrix):
        """
        Getting minimum diameter (mm) to be specified for pumps and valves
        """
        links_count = fn.network(d)[1]
        pumps_count = fn.network(d)[4]
        valves_count = fn.network(d)[5]
        pipe_diameter_matrix = fn.network(d)[16]
        assert fn.minimum_link_diameter(links_count, pumps_count, valves_count, pipe_diameter_matrix) == 203.2
        print('Function 5 test - Success')
        
    def test_sync_time(self, d, H, water_quality_time, hydraulic_steps_count, base_time_seconds, \
                       hydraulic_simulation_time, day_count, base_time_days, water_quality_step, \
                           expected_hydaulic_step, synchronising_option, reservoir_pattern):
        """
        Synchronizing water quality and hydraulic time steps
        """
        water_quality_time = 86700 # water quality time (s)
        hydraulic_steps_count = 27 # expected total time steps in the hydraulic report
        base_time_seconds = 86400 # base time period (s)
        hydraulic_simulation_time = 86400 # total hydraulic simuation time (s)
        day_count = 2 # count of the day in water quality simulation (86700 s corresponds to 2nd day)
        base_time_days = 1 # base time period (day)
        water_quality_step = 300 # water quality time step (s)
        expected_hydaulic_step = 25 # expected time step in the hydraulic report
        synchronising_option = 'steady'
        """Using dummy data for reservoir pattern"""
        reservoir_pattern = np.zeros((2, 288, 3))
        assert fn.sync_time(d, H, water_quality_time, hydraulic_steps_count, base_time_seconds, \
                       hydraulic_simulation_time, day_count, base_time_days, water_quality_step, \
                           expected_hydaulic_step, synchronising_option, reservoir_pattern) == [25, 25, 287, 287]
        print('Function 6 test - Success')
        
    def test_time_filter(self, H, ratio, hydraulic_steps_count, hydraulic_step):
        """
        Filtering out unwanted time steps from hydraulic analysis output
        """
        ratio = 1.2 # dummy value
        hydraulic_steps_count = 27 # expected total time steps in the hydraulic report
        hydraulic_step = 3600 # hydraulic time step (s)
        assert fn.time_filter(H, ratio, hydraulic_steps_count, hydraulic_step) == [5, 23]
        print('Function 7 test - Success')
        
    def test_time_data(self, time_matrix, filtered_time_step_matrix):
        """
        Cleaning time data
        """
        time_matrix = H.Time # from EPANET
        """Using dummy data for filtered_time_step_matrix"""
        filtered_time_step_matrix = [5, 23]
        assert (fn.time_data(time_matrix, filtered_time_step_matrix) == np.arange(0, 90000, 3600)).all()
        print('Function 8 test - Success')
        
    def test_velocity_data(self, pipe_velocity_matrix, filtered_time_step_matrix, flow_unit):
       """
       Cleaning velocity data
       """ 
       pipe_velocity_matrix = H.Velocity # from EPANET
       """Using dummy data for filtered_time_step_matrix"""
       filtered_time_step_matrix = [5, 23]
       flow_unit = fn.network(d)[12]
       assert (fn.velocity_data(pipe_velocity_matrix, filtered_time_step_matrix, flow_unit)[:, 118] == np.zeros(25)).all()
       print('Function 9 test - Success')
       
    def test_demand_data(self, node_demand_matrix, filtered_time_step_matrix, flow_unit):
       """
       Cleaning demand data
       """
       node_demand_matrix = H.Demand # from EPANET
       """Using dummy data for filtered_time_step_matrix"""
       filtered_time_step_matrix = [5, 23]
       flow_unit = fn.network(d)[12]
       assert (fn.demand_data(node_demand_matrix, filtered_time_step_matrix, flow_unit)[:, 2] == np.zeros(25)).all()
       print('Function 10 test - Success')
       
    def test_flow_data(self, pipe_flow_matrix, filtered_time_step_matrix, flow_unit):
        """
        Cleaning flow data
        """
        pipe_flow_matrix = H.Flow # from EPANET
        """Using dummy data for filtered_time_step_matrix"""
        filtered_time_step_matrix = [5, 23]
        flow_unit = fn.network(d)[12]
        assert (fn.flow_data(pipe_flow_matrix, filtered_time_step_matrix, flow_unit)[5: 22, 118] == np.zeros(17)).all()
        print('Function 11 test - Success')
        
    def test_tank_volume_data(self, tank_volume_matrix, filtered_time_step_matrix, flow_unit):
        """
        Cleaning tank volume data
        """
        tank_volume_matrix = H.TankVolume # from EPANET
        """Using dummy data for filtered_time_step_matrix"""
        filtered_time_step_matrix = [5, 23]
        flow_unit = fn.network(d)[12]
        assert (fn.tank_volume_data(tank_volume_matrix, filtered_time_step_matrix, flow_unit)[:, 94: 97] != 0).any()
        print('Function 12 test - Success')
        
    def test_maximum_segments(self, pipe_velocity_tolerance, water_quality_step, link_length_matrix, pipe_velocity_matrix):
        """
        Determining the maximum number of segments for pipe discretization
        """
        pipe_velocity_tolerance = 0.001 # tolerable flow velocity considered (m/s)
        water_quality_step = 300 # water quality time step (s)
        link_length_matrix = fn.network(d)[15]
        pipe_velocity_matrix = H.Velocity # from EPANET
        assert fn.maximum_segments(pipe_velocity_tolerance, water_quality_step, link_length_matrix, pipe_velocity_matrix) == 102
        print('Function 13 test - Success')
        
    def test_reservoir_names(self, reservoirs_count, node_names, reservoir_index_matrix):
        """
        Display the names of reservoirs
        """
        reservoirs_count = fn.network(d)[2]
        node_names = fn.network(d)[6]
        reservoir_index_matrix = fn.network(d)[8]
        assert fn.reservoir_names(reservoirs_count, node_names, reservoir_index_matrix) != ''
        print('Function 14 test - Success')
        
    def test_tank_names(self, tanks_count, node_names, tank_index_matrix):
        """
        Display the names of tanks
        """
        tanks_count = fn.network(d)[3]
        node_names = fn.network(d)[6]
        tank_index_matrix = fn.network(d)[9]
        assert fn.tank_names(tanks_count, node_names, tank_index_matrix) != ''
        print('Function 15 test - Success')
        
    def test_pump_names(self, pumps_count, valves_count, link_names):
        """
        Display the names of pumps
        """
        pumps_count = fn.network(d)[4]
        valves_count = fn.network(d)[5]
        link_names = fn.network(d)[7]
        assert fn.pump_names(pumps_count, valves_count, link_names) != ''
        print('Function 16 test - Success')
        
    def test_valve_names(self, valves_count, link_names):
        """
        Display the names of valves
        """
        valves_count = fn.network(d)[5]
        link_names = fn.network(d)[7]
        assert fn.valve_names(valves_count, link_names) == None
        print('Function 17 test - Success')
        
    def test_simulation_info(self, maximum_iterations_count, water_quality_time_days, water_quality_step, total_water_quality_steps):
        """
        Display simulation information
        """
        maximum_iterations_count = 200 # times for which the water qality simulation will be iteratively performed
        water_quality_time_days = 10 # days for which water quality will be simulated
        water_quality_step = 300 # water quality time step (s)
        total_water_quality_steps = 2881 # total water quality steps involved
        assert fn.simulation_info(maximum_iterations_count, water_quality_time_days, water_quality_step, total_water_quality_steps) != ''
        print('Function 18 test - Success')
        
    def test_msrt_info(self, quality_parameters_count, bulk_quality_parameters_count, wall_quality_parameters_count):
        """
        Display MSRT model information
        """
        """Using dummy data for quality_parameters_count"""
        quality_parameters_count = 3
        """Using dummy data for bulk_quality_parameters_count"""
        bulk_quality_parameters_count = 3
        """Using dummy data for wall_quality_parameters_count"""
        wall_quality_parameters_count = 0
        assert fn.msrt_info(quality_parameters_count, bulk_quality_parameters_count, wall_quality_parameters_count) != ''
        print('Function 19 test - Success')
        
    def test_omitted_nodes(self, omitted_nodes_count, node_names, omitted_node_index_matrix):
        """
        Display nodes omitted from analysis
        """
        omitted_nodes_count = fn.network(d)[19]
        node_names = fn.network(d)[6]
        omitted_node_index_matrix = fn.network(d)[20]
        assert fn.omitted_nodes(omitted_nodes_count, node_names, omitted_node_index_matrix) == None
        print('Function 20 test - Success')
        
    def test_omitted_links(self, omitted_links_count, link_names, omitted_link_index_matrix):
        """
        Display links omitted from analysis
        """
        omitted_links_count = fn.network(d)[21]
        link_names = fn.network(d)[7]
        omitted_link_index_matrix = fn.network(d)[22]
        assert fn.omitted_links(omitted_links_count, link_names, omitted_link_index_matrix) == None
        print('Function 21 test - Success')
        
test = TestEpytcFunctions()
test.test_network(d)
test.test_incoming_links(0, fn.network(d)[18])
test.test_outgoing_links(1, fn.network(d)[17])
test.test_minimum_link_length(27, 300, 0.001, H.Velocity)
test.test_minimum_link_diameter(fn.network(d)[1], fn.network(d)[4], fn.network(d)[5], fn.network(d)[16])
test.test_sync_time(d, H, 86700, 27, 86400,86400, 2, 1, 300, 25, 'steady', np.zeros((3, 288, 3)))
test.test_time_filter(H, 1.2, 27, 3600)
test.test_time_data(H.Time, [5, 23])
test.test_velocity_data(H.Velocity, [5, 23], 'GPM')
test.test_demand_data(H.Demand, [5, 23], 'GPM')
test.test_flow_data(H.Flow, [5, 23], 'GPM')
test.test_tank_volume_data(H.TankVolume, [5, 23], 'GPM')
test.test_maximum_segments(0.001, 300, fn.network(d)[15], H.Velocity)
test.test_reservoir_names(fn.network(d)[2], fn.network(d)[6], fn.network(d)[8])
test.test_tank_names(fn.network(d)[3], fn.network(d)[6], fn.network(d)[9])
test.test_pump_names(fn.network(d)[4], fn.network(d)[5], fn.network(d)[7])
test.test_valve_names(fn.network(d)[5], fn.network(d)[7])
test.test_simulation_info(200, 10, 300, 2881)
test.test_msrt_info(3, 3, 0)
test.test_omitted_nodes(fn.network(d)[19], fn.network(d)[6], fn.network(d)[20])
test.test_omitted_links(fn.network(d)[21], fn.network(d)[7], fn.network(d)[22])

print('\nAll tests completed!\n')




import os
from .run import run_epytc
from dataclasses import dataclass
from hydra.utils import instantiate
from omegaconf import OmegaConf


@dataclass
class epytc_class:
    module: str
    maximum_iterations_required: int
    simulation_period_days: int
    simulation_time_step: int
    base_period_days: int
    minimum_pipe_flow_velocity: float
    reservoir_quality_matrix: list[list[float]]
    reservoir_quality_pattern: str
    reservoir_quality_pattern_random_variability: float
    reservoir_injection_pattern: str
    reservoir_injection_pattern_random_variability: float
    reservoir_injection_start_time: list[list[float]]
    reservoir_injection_end_time: list[list[float]]
    reservoir_injection_input_value: list[list[float]]
    injection_nodes_index: list
    injection_nodes_quality_matrix: list[list[float]]
    injection_node_quality_pattern: str
    injection_node_quality_pattern_random_variability: float
    injection_node_injection_pattern: str
    injection_node_injection_pattern_random_variability: float
    injection_node_injection_start_time: list[list[float]]
    injection_node_injection_end_time: list[list[float]]
    injection_node_injection_input_value: list[list[float]]
    hyd_wq_sync_option: str


def create_epytc():
    """Creates an object of dataclass epytc-class

    :return: An epytc_class object
    :rtype: epytc_class
    """
    print("Creating instance of EPyT-C with default values")
    file_dir = os.path.dirname(os.path.abspath(__file__))
    default_config = OmegaConf.load(os.path.join(file_dir, "default_values.yaml"))
    epytc = instantiate(default_config.epytc_class)

    return epytc


def execute_epytc(epytc: epytc_class):
    """Execute the epytc module

    :param epytc: epytc object
    :type epytc: epytc_class
    """
    print("EPyT-C loaded for execution...")
    run_epytc(epytc)



import numpy as np
import math


class fn:
    def network(d):
        """Getting basic details of the network

        :param d: EPANET model
        :type d: EPANET object
        :return: Network details
        :rtype: List
        """
        network_info = [
            d.getNodeCount(),
            d.getLinkCount(),
            d.getNodeReservoirCount(),
            d.getNodeTankCount(),
            d.getLinkPumpCount(),
            d.getLinkValveCount(),
            d.getNodeNameID(),
            d.getLinkNameID(),
            d.getNodeReservoirIndex(),
            d.getNodeTankIndex(),
        ]
        index_pumps = []
        for x in range(
            d.getLinkCount() - (d.getLinkPumpCount() + d.getLinkValveCount()),
            len(d.getLinkNameID()) - d.getLinkValveCount(),
        ):
            index_pumps.append(x + 1)
        network_info.append(index_pumps)
        index_valves = []
        for x in range(
            d.getLinkCount() - d.getLinkValveCount(), len(d.getLinkNameID())
        ):
            index_valves.append(x + 1)
        network_info.append(index_valves)
        network_info.extend(
            [
                d.getFlowUnits(),
                d.getNodesConnectingLinksIndex(),
                d.getConnectivityMatrix(),
            ]
        )
        # Conversion of GPM units to SI units
        if d.getFlowUnits() == "GPM":
            network_info.append(0.3048 * d.getLinkLength())
            network_info.append(25.4 * d.getLinkDiameter())
        else:
            network_info.extend([d.getLinkLength(), d.getLinkDiameter()])
        start_node_matrix = []
        end_node_matrix = []
        for x in range(d.getLinkCount()):
            var1 = d.getNodesConnectingLinksIndex()[x][0]
            var2 = d.getNodesConnectingLinksIndex()[x][1]
            start_node_matrix.append(var1)
            end_node_matrix.append(var2)
        network_info.extend([start_node_matrix, end_node_matrix])
        # Number of nodes omitted for analysis, if any
        number_omitted_nodes = 0
        # Index of omitted nodes
        index_omitted_nodes = []
        # Number of links omitted for analysis, if any
        number_omitted_links = 0
        # Index of omitted nodes
        index_omitted_links = []
        network_info.extend(
            [
                number_omitted_nodes,
                index_omitted_nodes,
                number_omitted_links,
                index_omitted_links,
            ]
        )
        return network_info

    def incoming_links(num1, arr1):
        """Getting incoming links to a specific node

        :param num1: ID of a node
        :type num1: Integer
        :param arr1: Matrix of end node indices
        :type arr1: List
        :return: List of incoming links
        :rtype: List
        """
        links_connecting_to_node = []
        if arr1.count(num1 + 1) > 0:
            arr = np.array(arr1)
            bool_arr = arr == num1 + 1
            links_connecting_to_node = np.where(bool_arr)[0]
        return links_connecting_to_node

    def outgoing_links(num1, arr1):
        """Getting outgoing links from a specific node

        :param num1: ID of a node
        :type num1: Integer
        :param arr1: Matrix of start node indices
        :type arr1: List
        :return: List of outgoing links
        :rtype: List
        """
        links_connecting_from_node = []
        if arr1.count(num1 + 1) > 0:
            arr = np.array(arr1)
            bool_arr = arr == num1 + 1
            links_connecting_from_node = np.where(bool_arr)[0]
        return links_connecting_from_node

    def minimum_link_length(num1, num2, num3, arr1):
        """Estimating the minimum length to be specified for pumps and valves

        :param num1: Time steps in the filtered hydraulic report
        :type num1: Integer
        :param num2: Water quality time step
        :type num2: Integer
        :param num3: Tolerable flow velocity considered
        :type num3: Integer
        :param arr1: Matrix of link flow velocities
        :type arr1: List
        :return: Minimum link length
        :rtype: Float
        """
        min_vel = []
        for h in range(num1):
            min_val = np.min(arr1[h])
            min_vel.append(min_val)
        minimum_flow_velocity = np.min(min_vel)
        if minimum_flow_velocity < num3:
            minimum_flow_velocity = num3
        minimum_link_length = minimum_flow_velocity * num2
        return minimum_link_length

    def minimum_link_diameter(num1, num2, num3, arr1):
        """Estimating the minimum diameter to be specified for pumps and valves

        :param num1: Count of links
        :type num1: Integer
        :param num2: Count of pumps
        :type num2: Integer
        :param num3: Count of valves
        :type num3: Integer
        :param arr1: Matrix of link diameters
        :type arr1: List
        :return: Minimum diameter
        :rtype: Float
        """
        min_val = np.min(arr1)
        if min_val == 0:
            arr2 = arr1[0 : num1 - (num2 + num3)]
            min_val = np.min(arr2)
            minimum_link_diameter = min_val
        else:
            minimum_link_diameter = min_val
        return minimum_link_diameter

    def sync_time(d, H, num1, num2, num3, num4, num5, num6, num7, num8, str1, arr2):
        """Synchronizing water quality and hydraulic time steps

        :param d: EPANET model
        :type d: EPANET object
        :param H: Hydraulic simulation output from EPANET
        :type H: List
        :param num1: Current water quality time in the simulation in seconds
        :type num1: Integer
        :param num2: Total steps in the hydraulic report 'Time' that was expected
        :type num2: Integer
        :param num3: Base time period in seconds
        :type num3: Integer
        :param num4: Total hydraulic simuation time in seconds
        :type num4: Integer
        :param num5: Count of the day in water quality simulation
        :type num5: Integer
        :param num6: Base time period in day(s)
        :type num6: Float/ Integer
        :param num7: Water quality simulation time step in seconds
        :type num7: Integer
        :param num8: Expected time step in the hydraulic report
        :type num8: Integer
        :param str1: Option for synchronization
        :type str1: String
        :param arr1: Input pattern governing quality values(input) at the reservoir
        :type str1: Array
        :return: Synchronized steps
        :rtype: List
        """
        if str1 == "steady":
            if num1 == 0:
                h_step_expected = math.floor(num2 - (num3 / num4)) - 1
                h_step = h_step_expected
                reservoir_pattern_step = injection_pattern_step = 0
            else:
                h_step_expected = num8
                wq_time_cycle = num1 - (num5 - 1) * num6 * 24 * 3600
                wq_time_hydraulic_report = wq_time_cycle + (
                    int(d.getTimeSimulationDuration()) - num3
                )
                for x in range(num8, num2):
                    if wq_time_hydraulic_report <= H.Time[x]:
                        h_step = math.floor(x)
                        break
                reservoir_pattern_step = injection_pattern_step = (
                    len(arr2[0]) - (num2 - h_step) + 1
                )
        elif str1 == "dynamic":
            if num1 == 0:
                h_step_expected = 0
                h_step = h_step_expected
                reservoir_pattern_step = injection_pattern_step = 0
            else:
                h_step_expected = num8
                wq_time_cycle = num1
                wq_time_hydraulic_report = wq_time_cycle
                for x in range(num8, num2):
                    if wq_time_hydraulic_report <= H.Time[x]:
                        h_step = math.floor(x)
                        break
                reservoir_pattern_step = injection_pattern_step = h_step - int(
                    (num5 - 1) * (num6 * 24 * 3600 / num7)
                )
        out = [h_step, h_step_expected, reservoir_pattern_step, injection_pattern_step]
        return out

    def time_filter(H, num1, num2, num3):
        """Filtering out unwanted time steps from hydraulic analysis output

        :param H: Hydraulic simulation output from EPANET
        :type H: List
        :param num1: Ratio of time steps actually reported to time steps expected in the report
        :type num1: Float
        :param num2: Total steps in the hydraulic report 'Time' that was expected
        :type num2: Integer
        :param num3: Hydraulic simulation time step in seconds
        :type num3: Integer
        :return: Details of unwanted steps for water quality simulation
        :rtype: List
        """
        filtered_steps = []
        if num1 > 1:
            print("Filtering hydraulic analysis report.")
            for s in range(num2):
                if H.Time[s] % num3 != 0:
                    filtered_steps.append(s)
        return filtered_steps

    def time_data(arr1, arr2):
        """Cleaning time data

        :param arr1: Time ouput from EPANET hydraulic simulation
        :type arr1: List
        :param arr2: Details of unwanted steps for water quality simulation
        :return: Cleansed time output
        :rtype: List
        """
        arr2 = np.array(arr2)
        if len(arr2) > 0:
            out = np.delete(arr1, arr2, 0)
        else:
            out = arr1
        return out

    def velocity_data(arr1, arr2, str1):
        """Cleaning velocity data

        :param arr1: Link velocity ouput from EPANET hydraulic simulation
        :type arr1: List
        :param arr2: Details of unwanted steps for water quality simulation
        :return: Cleansed time output
        :param str1: Flow Unit
        :type str1: String
        :return: Cleansed velocity output
        :rtype: List
        """
        arr2 = np.array(arr2)
        if len(arr2) > 0:
            out = np.delete(arr1, arr2, 0)
        else:
            out = arr1
        if str1 == "GPM":
            out = np.multiply(out, 0.3048)
        return out

    def demand_data(arr1, arr2, str1):
        """Cleaning demand data

        :param arr1: Node demand ouput from EPANET hydraulic simulation
        :type arr1: List
        :param arr2: Details of unwanted steps for water quality simulation
        :return: Cleansed time output
        :param str1: Flow Unit
        :type str1: String
        :return: Cleansed demand output
        :rtype: List
        """
        arr2 = np.array(arr2)
        if len(arr2) > 0:
            out = np.delete(arr1, arr2, 0)
        else:
            out = arr1
        if str1 == "GPM":
            out = np.multiply(out, 6.3e-5)
        elif str1 == "LPS":
            out = np.multiply(out, 1e-3)
        elif str1 == "LPM":
            out = np.multiply(out, 1.67e-5)
        elif str1 == "CMH":
            out = np.multiply(out, (1 / 3600))
        return out

    def flow_data(arr1, arr2, str1):
        """Cleaning flow data

        :param arr1: Link flow ouput from EPANET hydraulic simulation
        :type arr1: List
        :param arr2: Details of unwanted steps for water quality simulation
        :return: Cleansed time output
        :param str1: Flow Unit
        :type str1: String
        :return: Cleansed flow output
        :rtype: List
        """
        arr2 = np.array(arr2)
        if len(arr2) > 0:
            out = np.delete(arr1, arr2, 0)
        else:
            out = arr1
        if str1 == "GPM":
            out = np.multiply(out, 6.3e-5)
        elif str1 == "LPS":
            out = np.multiply(out, 1e-3)
        elif str1 == "LPM":
            out = np.multiply(out, 1.67e-5)
        elif str1 == "CMH":
            out = np.multiply(out, (1 / 3600))
        return out

    def tank_volume_data(arr1, arr2, str1):
        """Cleaning tank volume data

        :param arr1: Tank volume ouput from EPANET hydraulic simulation
        :type arr1: List
        :param arr2: Details of unwanted steps for water quality simulation
        :return: Cleansed time output
        :param str1: Flow Unit
        :type str1: String
        :return: Cleansed tank volume output
        :rtype: List
        """
        arr2 = np.array(arr2)
        if len(arr2) > 0:
            out = np.delete(arr1, arr2, 0)
        else:
            out = arr1
        if str1 == "GPM":
            out = np.multiply(out, 0.0283)
        return out

    def maximum_segments(num1, num2, arr1, arr2):
        """Determining the maximum number of segments for pipe discretization

        :param num1: Tolerable flow velocity considered in metres per second
        :type num1: Float
        :param num2: Water quality simulation time step in seconds
        :type num2: Integer
        :param arr1: Matrix of link lengths
        :type arr1: List
        :param arr2: Matrix of link velocity values
        :type arr2: List
        :return: Maximum number of pipe segments
        :rtype: Integer
        """
        arr2[arr2 < num1] = 0
        arr2[arr2 == 0] = 100
        min_vel = np.min(arr2)
        index = np.where(arr2 == min_vel)
        len_min_vel_pipe = arr1[index[1][0]]
        max_segments = math.ceil(len_min_vel_pipe / (num1 * num2)) + 1
        return max_segments

    def reservoir_names(num1, arr1, arr2):
        """Displaying the names of reservoirs

        :param num1: Count of reservoirs
        :type num1: Integer
        :param arr1: Matrix of node names
        :type arr1: List
        :param arr2: Matrix of reservoir indices
        :type arr2: List
        :return: Names of reservoirs
        :rtype: String
        """
        for x in range(num1):
            print("Reservoir %d: %s" % (x + 1, arr1[arr2[x] - 1]))

    def tank_names(num1, arr1, arr2):
        """Displaying the names of tanks

        :param num1: Count of tanks
        :type num1: Integer
        :param arr1: Matrix of node names
        :type arr1: List
        :param arr2: Matrix of tank indices
        :type arr2: List
        :return: Names of tanks
        :rtype: String
        """
        for x in range(num1):
            print("Tank %d: %s" % (x + 1, arr1[arr2[x] - 1]))

    def pump_names(num1, num2, arr1):
        """Displaying the names of pumps

        :param num1: Count of pumps
        :type num1: Integer
        :param num2: Count of valves
        :type num2: Integer
        :param arr1: Matrix of link names
        :type arr1: List
        :return: Names of pumps
        :rtype: String
        """
        if num1 > 0:
            for x in range(num1):
                print("Pump %d: %s" % (x + 1, arr1[len(arr1) - num2 - (num1 - x)]))

    def valve_names(num1, arr1):
        """Displaying the names of valves

        :param num1: Count of valves
        :type num1: Integer
        :param arr1: Matrix of link names
        :type arr1: List
        :return: Names of valves
        :rtype: String
        """
        if num1 > 0:
            for x in range(num1):
                print("Valve %d: %s" % (x + 1, arr1[len(arr1) - (num1 - x)]))

    def simulation_info(num1, num2, num3, num4):
        """Displaying the basic information about the inputs selected for the simulation

        :param num1: Maximum number of iterations
        :type num1: Integer
        :param num2: Water quality simuation time in days
        :type num2: Float/ Integer
        :param num3: Water quality simuation time step in seconds
        :type num3: Integer
        :param num4: Total number of water quality simulation steps
        :type num4: Integer
        :return: Basic information about simulation
        :rtype: String
        """
        print("Number of iterations: %d" % (num1))
        print("Number of days for which water quality is simulated: %d" % (num2))
        print("Water quality simulation time step: %d seconds" % (num3))
        print("Number of water quality simulation steps: %d" % (num4))

    def msrt_info(num1, num2, num3):
        """Displaying the basic information about the MSRT model selected for the simulation

        :param num1: Count of water quality parameters
        :type num1: Integer
        :param num2: Count of bulk phase water quality parameters
        :type num2: Float/ Integer
        :param num3: Count of wall phase water quality parameters
        :type num3: Integer
        :return: Basic information about MSRT model
        :rtype: String
        """
        print("Number of water quality parameters in the MSRT model: %d" % (num1))
        print("Number of bulk phase water quality paraneters: %d" % (num2))
        print("Number of wall phase water quality paraneters: %d" % (num3))

    def omitted_nodes(num1, arr1, arr2):
        """Displaying the information about the nodes (if any) omitted from water quality analysis

        :param num1: Count of omitted nodes
        :type num1: Integer
        :param arr1: Matrix of node names
        :type arr1: List
        :param arr2: Matrix of indices of omitted nodes
        :type arr2: List
        :return: Details of omitted nodes
        :rtype: String
        """
        if num1 > 0:
            print("Number of nodes omitted for analysis: %d" % (num1))
            for x in range(num1):
                print("Omitted Node %d: %s" % (x + 1, arr1[arr2[x]]))

    def omitted_links(num1, arr1, arr2):
        """Displaying the information about the links (if any) omitted from water quality analysis

        :param num1: Count of omitted links
        :type num1: Integer
        :param arr1: Matrix of link names
        :type arr1: List
        :param arr2: Matrix of indices of omitted links
        :type arr2: List
        :return: Details of omitted links
        :rtype: String
        """
        if num1 > 0:
            print("Number of links omitted for analysis: %d" % (num1))
            for x in range(num1):
                print("Omitted Link %d: %s" % (x + 1, arr1[arr2[x]]))



from sys import exit
from .functions import fn
import os
from os import getcwd
import math
import numpy as np
import pandas as pd
import copy
import time
from epyt import epanet

def run_epytc(epytc):
    """Runs the simulation usign the data from the epytc object

    :param epytc: epytc object
    :type epytc: epytc_class
    """
    if epytc.module == "MSRT-1":
        from .chlorine_decay_thms_formation import module
    elif epytc.module == "MSRT-2":
        from .bacterial_regrowth import module
    elif epytc.module == "MSRT-3":
        from .arsenite_oxidation_arsenate_attachment_detachment import module
    elif epytc.module == "MSRT-4":
        from .pfas_formation import module

    start_time = time.time()
    # Display MSRT module details
    module.details()
    # Input data
    wq_max_iteration = epytc.maximum_iterations_required
    wq_sim_days = epytc.simulation_period_days
    wq_sim_time_step_s = epytc.simulation_time_step
    base_time_cycle_day = epytc.base_period_days
    total_wq_steps = int(((wq_sim_days * 24 * 3600) / wq_sim_time_step_s) + 1)
    base_time_cycle_s = base_time_cycle_day * 24 * 3600
    Tolerable_u = epytc.minimum_pipe_flow_velocity
    reservoir_quality_input = epytc.reservoir_quality_matrix
    reservoir_quality_pattern = epytc.reservoir_quality_pattern
    reservoir_quality_pattern_rand_var = (
        epytc.reservoir_quality_pattern_random_variability
    )
    reservoir_injection_pattern = epytc.reservoir_injection_pattern
    reservoir_injection_pattern_rand_var = (
        epytc.reservoir_injection_pattern_random_variability
    )
    reservoir_injection_start_time = epytc.reservoir_injection_start_time
    reservoir_injection_end_time = epytc.reservoir_injection_end_time
    reservoir_injection_input_val = epytc.reservoir_injection_input_value
    index_injection_nodes = epytc.injection_nodes_index
    injection_nodes_quality_mat = epytc.injection_nodes_quality_matrix
    injection_node_quality_pattern = epytc.injection_node_quality_pattern
    injection_node_quality_pattern_rand_var = (
        epytc.injection_node_quality_pattern_random_variability
    )
    injection_node_injection_pattern = epytc.injection_node_injection_pattern
    injection_node_injection_pattern_rand_var = (
        epytc.injection_node_injection_pattern_random_variability
    )
    injection_node_injection_start_time = epytc.injection_node_injection_start_time
    injection_node_injection_end_time = epytc.injection_node_injection_end_time
    injection_node_injection_input_val = epytc.injection_node_injection_input_value
    sync_option = epytc.hyd_wq_sync_option
    # Load network
    d = epanet(epytc.network_name)
    # Creating variables
    num_nodes = fn.network(d)[0]
    num_links = fn.network(d)[1]
    num_reservoirs = fn.network(d)[2]
    num_tanks = fn.network(d)[3]
    num_pumps = fn.network(d)[4]
    num_valves = fn.network(d)[5]
    name_nodes = fn.network(d)[6]
    name_links = fn.network(d)[7]
    index_reservoirs = fn.network(d)[8]
    index_tanks = fn.network(d)[9]
    index_pumps = fn.network(d)[10]
    index_valves = fn.network(d)[11]
    unit = fn.network(d)[12]
    link_connectivity_array = fn.network(d)[13]
    connectivity_mat = fn.network(d)[14]
    length_links = fn.network(d)[15]
    diameter_links = fn.network(d)[16]
    start_node_mat = fn.network(d)[17]
    end_node_mat = fn.network(d)[18]
    num_omitted_nodes = fn.network(d)[19]
    index_omitted_nodes = fn.network(d)[20]
    num_omitted_links = fn.network(d)[21]
    index_omitted_links = fn.network(d)[22]
    wq_parameter_num = module.species()[0]
    bulk_wq_parameter_num = module.species()[1]
    wall_wq_parameter_num = module.species()[2]
    num_variables = module.species()[3]
    # Displaying relevant information
    fn.reservoir_names(num_reservoirs, name_nodes, index_reservoirs)
    fn.tank_names(num_tanks, name_nodes, index_tanks)
    fn.pump_names(num_pumps, num_valves, name_links)
    fn.valve_names(num_valves, name_links)
    fn.simulation_info(
        wq_max_iteration, wq_sim_days, wq_sim_time_step_s, total_wq_steps
    )
    fn.msrt_info(wq_parameter_num, bulk_wq_parameter_num, wall_wq_parameter_num)
    fn.omitted_nodes(num_omitted_nodes, name_nodes, index_omitted_nodes)
    fn.omitted_links(num_omitted_links, name_links, index_omitted_links)
    # Getting the values of MSRT module variables
    variable_values_mat = module.variables(wq_max_iteration, num_variables)
    # Getting the quality characteristics of reservoir(s)
    reservoir_quality = module.reservoir_quality(
        d,
        wq_max_iteration,
        reservoir_quality_input,
        reservoir_quality_pattern,
        reservoir_quality_pattern_rand_var,
    )
    # Getting the quality characteristics of reservoir(s)
    injection_quality = module.injection_quality(
        wq_max_iteration,
        index_injection_nodes,
        injection_nodes_quality_mat,
        injection_node_quality_pattern,
        injection_node_quality_pattern_rand_var,
    )
    wq_iteration_count = 1
    # Start of water quality simulation
    while wq_iteration_count <= wq_max_iteration:
        print(
            "Water quality simulation (Iteration %d) is starting..."
            % (wq_iteration_count)
        )
        # Hydraulic simulation Module
        # Number of seconds for which hydraulics is simulated
        h_sim_s = d.getTimeSimulationDuration()
        if h_sim_s < (wq_sim_days * 24 * 3600):
            d.setTimeSimulationDuration(wq_sim_days * 24 * 3600)
            print(
                "Hydraulic analysis simulation period (changed to): %d days"
                % (wq_sim_days)
            )
        else:
            print(
                "Hydraulic analysis simulation period: %d days"
                % (h_sim_s / (24 * 3600))
            )
        # Storing hydraulic time step in seconds
        h_sim_time_step_s = d.getTimeHydraulicStep()
        print("Time period for hydraulic analysis: %d seconds" % (h_sim_time_step_s))
        # EPANET analysis
        d.setQualityType("age")
        H = d.getComputedHydraulicTimeSeries()
        Q = d.getComputedQualityTimeSeries()
        print("Analysis with EPANET completed...")
        print("Information successfully stored.")
        # Interpreting hydraulic analysis report
        total_h_steps_reported = len(H.Time)
        total_h_steps_expected = int(
            (d.getTimeSimulationDuration() / h_sim_time_step_s) + 1
        )
        h_ratio_report = (total_h_steps_reported - 1) / (total_h_steps_expected - 1)
        # Filter uneven timesteps
        filter_steps_mat = fn.time_filter(
            H, h_ratio_report, total_h_steps_reported, h_sim_time_step_s
        )
        H.Time = fn.time_data(H.Time, filter_steps_mat)
        H.Velocity = fn.velocity_data(H.Velocity, filter_steps_mat, unit)
        H.Demand = fn.demand_data(H.Demand, filter_steps_mat, unit)
        H.Flow = fn.flow_data(H.Flow, filter_steps_mat, unit)
        H.TankVolume = fn.tank_volume_data(H.TankVolume, filter_steps_mat, unit)
        # Creating temporary storage matrices
        link_conc_data_points_mat = np.zeros((total_wq_steps, num_links))
        link_segments_width_mat = np.zeros((total_wq_steps, num_links))
        link_flow_velocity_mat = np.zeros((total_wq_steps, num_links))
        link_flow_Reynolds_num_mat = np.zeros((total_wq_steps, num_links))
        tank_flow_volume_mat = np.zeros((total_wq_steps, num_tanks))
        # Getting the maximum number of link grid points required
        link_conc_data_points_max = fn.maximum_segments(
            Tolerable_u, wq_sim_time_step_s, length_links, H.Velocity
        )
        link_conc_array_a = np.zeros(
            (wq_parameter_num, link_conc_data_points_max, num_links)
        )
        link_conc_array = np.zeros(
            (wq_parameter_num, link_conc_data_points_max, num_links)
        )
        node_conc_array_a = np.zeros((bulk_wq_parameter_num, total_wq_steps, num_nodes))
        node_conc_array = np.zeros((bulk_wq_parameter_num, total_wq_steps, num_nodes))
        # Reservoir water quality
        reservoir_quality_mat = np.zeros((num_reservoirs, bulk_wq_parameter_num))
        for x in range(num_reservoirs):
            reservoir_quality_mat[x] = reservoir_quality[x][wq_iteration_count - 1]
        # Injection node(s) water quality
        injection_quality_mat = np.zeros(
            (len(index_injection_nodes), bulk_wq_parameter_num)
        )
        for x in range(len(index_injection_nodes)):
            injection_quality_mat[x] = injection_quality[x][wq_iteration_count - 1]
        # Storing the initial condition for all nodes
        node_conc_initial_mat = np.zeros((num_nodes, bulk_wq_parameter_num))
        for x in range(num_reservoirs):
            node_conc_initial_mat[index_reservoirs[x] - 1] = reservoir_quality_mat[x]
        for x in range(len(index_injection_nodes)):
            for sp in range(bulk_wq_parameter_num):
                if injection_quality[x][wq_iteration_count - 1][sp] != 0:
                    node_conc_initial_mat[index_injection_nodes[x] - 1][sp] = (
                        injection_quality[x][wq_iteration_count - 1][sp]
                    )
        # Storing the initial condition for all links
        link_conc_initial_mat = np.zeros((num_links, wall_wq_parameter_num))
        print("Water quality simulation started...")
        wq_time = 0
        wq_step = 0
        reservoir_pattern = module.reservoir_pattern(
            d,
            base_time_cycle_day,
            reservoir_injection_pattern,
            reservoir_injection_pattern_rand_var,
            reservoir_injection_start_time,
            reservoir_injection_end_time,
            reservoir_injection_input_val,
        )
        injection_pattern = module.injection_pattern(
            d,
            base_time_cycle_day,
            index_injection_nodes,
            injection_node_injection_pattern,
            injection_node_injection_pattern_rand_var,
            injection_node_injection_start_time,
            injection_node_injection_end_time,
            injection_node_injection_input_val,
        )
        while wq_time <= wq_sim_days * 24 * 3600:
            print(
                "Water quality simulation step (Iteration {}/ {}): {} (/ {})".format(
                    wq_iteration_count, wq_max_iteration, wq_step + 1, total_wq_steps
                )
            )
            wq_step += 1
            time_cycle_count = math.ceil(
                (wq_time / (base_time_cycle_day * 24 * 3600)) / 1
            )

            # Synchronizing hydraulic and water quality information
            if wq_time == 0:
                h_step_expected = 0
            sync_out = fn.sync_time(
                d,
                H,
                wq_time,
                total_h_steps_expected,
                base_time_cycle_s,
                h_sim_time_step_s,
                time_cycle_count,
                base_time_cycle_day,
                wq_sim_time_step_s,
                h_step_expected,
                sync_option,
                reservoir_pattern,
            )
            h_step = sync_out[0]
            h_step_expected = sync_out[1]
            reservoir_pattern_step = sync_out[2]
            injection_pattern_step = sync_out[3]
            # Lagrangian stage
            count_links_lagrangian = 0
            link_flow_velocity_time_step = H.Velocity[h_step]
            link_flow_rate_time_step = H.Flow[h_step]
            link_flow_rate_prev_time_step = H.Flow[h_step - 1]
            for p in range(num_links):
                if index_omitted_links.count(p + 1) > 0:
                    count_links_lagrangian += 0
                else:
                    count_links_lagrangian += 1
                    nodes_connecting_link = d.getLinkNodesIndex(p + 1)
                    # Getting the flow velocity of link
                    link_flow_velocity = abs(link_flow_velocity_time_step[p])
                    # Getting the flow rate inside link
                    link_flow_rate = link_flow_rate_time_step[p]
                    # Getting the flow rate inside link for previous time step
                    link_flow_rate_prev = link_flow_rate_prev_time_step[p]

                    if index_pumps.count(p + 1) > 0 or index_valves.count(p + 1) > 0:
                        length_links[p] = fn.minimum_link_length(
                            total_h_steps_expected,
                            wq_sim_time_step_s,
                            Tolerable_u,
                            H.Velocity,
                        )
                        diameter_links[p] = fn.minimum_link_diameter(
                            num_links, num_pumps, num_valves, diameter_links
                        )
                    elif link_flow_velocity > 0 and link_flow_velocity < Tolerable_u:
                        link_flow_velocity = 0
                    if length_links[p] != 0:
                        # Getting the number of segments into which a link is divided
                        if link_flow_velocity > 0:
                            link_segments_num = math.floor(
                                length_links[p]
                                / (link_flow_velocity * wq_sim_time_step_s)
                            )
                        else:
                            link_segments_num = 0
                        if link_segments_num <= 1:
                            link_segments_num = 2
                        elif link_segments_num > (link_conc_data_points_max - 1):
                            link_segments_num = link_conc_data_points_max - 1
                        # Getting the width of the segments into which a link is divided
                        link_segments_width = length_links[p] / link_segments_num
                        link_segments_width_mat[wq_step - 1][p] = link_segments_width
                        link_conc_data_points_mat[wq_step - 1][p] = (
                            link_segments_num + 1
                        )
                        # Checking for reservoir connected to a link
                        start_node = nodes_connecting_link[0]
                        if start_node in index_reservoirs:
                            for sp in range(bulk_wq_parameter_num):
                                if wq_step == 1:
                                    link_conc_array_a[sp][0][p] = reservoir_quality_mat[
                                        index_reservoirs.index(start_node)
                                    ][sp]
                                else:
                                    link_conc_array_a[sp][0][p] = (
                                        reservoir_quality_mat[
                                            index_reservoirs.index(start_node)
                                        ][sp]
                                        * reservoir_pattern[
                                            index_reservoirs.index(start_node)
                                        ][reservoir_pattern_step - 1][sp]
                                    )
                        elif start_node in index_injection_nodes:
                            for sp in range(bulk_wq_parameter_num):
                                if (
                                    injection_quality_mat[
                                        index_injection_nodes.index(start_node)
                                    ][sp]
                                    != 0
                                ):
                                    if wq_step == 1:
                                        link_conc_array_a[sp][0][p] = (
                                            injection_quality_mat[
                                                index_injection_nodes.index(start_node)
                                            ][sp]
                                        )
                                    else:
                                        link_conc_array_a[sp][0][p] = (
                                            injection_quality_mat[
                                                index_injection_nodes.index(start_node)
                                            ][sp]
                                            * injection_pattern[
                                                index_injection_nodes.index(start_node)
                                            ][injection_pattern_step - 1][sp]
                                        )
                        # Calculating the distance of grid points from rear end of a link
                        segment_time_step_mat = np.zeros((link_segments_num + 1, 1))
                        segment_prev_time_step_mat = np.zeros(
                            (int(link_conc_data_points_mat[wq_step - 2][p]), 1)
                        )
                        for i in range(1, link_segments_num + 1):
                            segment_time_step_mat[i][0] = (
                                segment_time_step_mat[i - 1][0] + link_segments_width
                            )

                        # Calculating the distance of grid points from rear end of a link for previous time step
                        if wq_step == 1:
                            segment_prev_time_step_mat = segment_time_step_mat
                        else:
                            for i in range(
                                1, int(link_conc_data_points_mat[wq_step - 2][p])
                            ):
                                segment_prev_time_step_mat[i][0] = (
                                    segment_prev_time_step_mat[i - 1][0]
                                    + link_segments_width_mat[wq_step - 2][p]
                                )
                        # Application of method of characteristics
                        alpha_mat = np.zeros((link_segments_num + 1, 1))
                        if (
                            wq_step == 1
                            or h_step == h_step_expected
                            or np.sign(link_flow_rate) == np.sign(link_flow_rate_prev)
                        ):
                            first_grid = 1
                            last_grid = link_segments_num + 1
                        else:
                            first_grid = 0
                            last_grid = link_segments_num
                        for i in range(first_grid, last_grid):
                            if (
                                wq_step == 1
                                or h_step == h_step_expected
                                or np.sign(link_flow_rate)
                                == np.sign(link_flow_rate_prev)
                            ):
                                alpha_mat[i][0] = segment_time_step_mat[i][0] - (
                                    link_flow_velocity * wq_sim_time_step_s
                                )
                            else:
                                alpha_mat[i][0] = segment_time_step_mat[i][0] + (
                                    link_flow_velocity * wq_sim_time_step_s
                                )
                            if alpha_mat[i][0] < 0:
                                alpha_mat[i][0] = 0
                            elif alpha_mat[i][0] > length_links[p]:
                                alpha_mat[i][0] = length_links[p]
                            if wq_step == 1:
                                ratio_segment = alpha_mat[i][0] / link_segments_width
                            else:
                                ratio_segment = (
                                    alpha_mat[i][0]
                                    / link_segments_width_mat[wq_step - 2][p]
                                )
                            prev_grid = math.floor(ratio_segment)
                            if prev_grid <= 0:
                                prev_grid = 0
                            next_grid = math.ceil(ratio_segment)
                            if next_grid >= link_segments_num + 1:
                                next_grid = link_segments_num + 1
                            elif next_grid == 0:
                                next_grid = 1
                            for sp in range(bulk_wq_parameter_num):
                                if prev_grid == next_grid:
                                    if (
                                        index_pumps.count(p + 1) > 0
                                        or index_valves.count(p + 1) > 0
                                    ):
                                        link_conc_array_a[sp][i][p] = link_conc_array_a[
                                            sp
                                        ][0][p]
                                    else:
                                        prev_grid -= 1
                                        link_conc_array_a[sp][i][p] = link_conc_array[
                                            sp
                                        ][prev_grid][p]
                                elif wq_step == 1:
                                    link_conc_array_a[sp][i][p] = link_conc_array[sp][
                                        prev_grid
                                    ][p] + (
                                        (
                                            link_conc_array[sp][next_grid][p]
                                            - link_conc_array[sp][prev_grid][p]
                                        )
                                        / link_segments_width
                                    ) * (
                                        alpha_mat[i][0]
                                        - segment_prev_time_step_mat[prev_grid][0]
                                    )
                                else:
                                    link_conc_array_a[sp][i][p] = link_conc_array[sp][
                                        prev_grid
                                    ][p] + (
                                        (
                                            link_conc_array[sp][next_grid][p]
                                            - link_conc_array[sp][prev_grid][p]
                                        )
                                        / link_segments_width_mat[wq_step - 2][p]
                                    ) * (
                                        alpha_mat[i][0]
                                        - segment_prev_time_step_mat[prev_grid][0]
                                    )
                            if (
                                index_pumps.count(p + 1) == 0
                                and index_valves.count(p + 1) == 0
                            ):
                                delta = module.pipe_reaction(
                                    wq_sim_time_step_s,
                                    p,
                                    i,
                                    link_flow_velocity,
                                    diameter_links[p],
                                    length_links[p],
                                    link_segments_width,
                                    variable_values_mat[wq_iteration_count - 1],
                                    link_conc_array_a,
                                )
                                for sp in range(wq_parameter_num):
                                    link_conc_array_a[sp][i][p] = (
                                        link_conc_array_a[sp][i][p] + delta[sp]
                                    )
                        link_conc_array_a[link_conc_array_a < 0] = 0
            if count_links_lagrangian != (num_links - num_omitted_links):
                print("Error in Lagrangian stage dedicated to pipes!")
                exit()
            count_nodes_lagrangian = 0
            node_demand_time_step = H.Demand[h_step]
            tank_flow_volume_time_step = H.TankVolume[h_step]
            for n in range(num_nodes):
                if index_omitted_nodes.count(n + 1) > 0:
                    count_nodes_lagrangian += 0
                else:
                    count_nodes_lagrangian += 1
                    flag_reservoir = 0
                    # Getting the demand of node
                    node_demand = node_demand_time_step[n]

                    if index_reservoirs.count(n + 1) > 0:
                        # Checking for reservoir
                        flag_reservoir = 1
                        for sp in range(bulk_wq_parameter_num):
                            if wq_step == 1:
                                node_conc_array_a[sp][wq_step - 1][n] = (
                                    reservoir_quality_mat[
                                        index_reservoirs.index(n + 1)
                                    ][sp]
                                )
                            else:
                                node_conc_array_a[sp][wq_step - 1][n] = (
                                    reservoir_quality_mat[
                                        index_reservoirs.index(n + 1)
                                    ][sp]
                                    * reservoir_pattern[index_reservoirs.index(n + 1)][
                                        reservoir_pattern_step - 1
                                    ][sp]
                                )
                    elif index_tanks.count(n + 1) > 0:
                        # Checking for tank
                        t = index_tanks.index(n + 1)
                        # Getting tank volume
                        tank_flow_volume = tank_flow_volume_time_step[n]
                        tank_flow_volume_mat[wq_step - 1][t] = tank_flow_volume
                        tank_flow_volume_prev = tank_flow_volume_mat[wq_step - 2][t]
                        links_connecting_to_node = fn.incoming_links(n, end_node_mat)
                        links_connecting_from_node = fn.outgoing_links(
                            n, start_node_mat
                        )
                        num_incoming_links = len(links_connecting_to_node)
                        num_outgoing_links = len(links_connecting_from_node)
                        mass_incoming_mat = np.zeros((bulk_wq_parameter_num, 1))
                        tank_outflow = 0
                        for i in range(num_incoming_links):
                            incoming_link = links_connecting_to_node[i]
                            if index_omitted_links.count(incoming_link + 1) > 0:
                                break
                            else:
                                link_flow_rate = link_flow_rate_time_step[incoming_link]
                                if link_flow_rate >= 0:
                                    pos = int(
                                        link_conc_data_points_mat[wq_step - 1][
                                            incoming_link
                                        ]
                                        - 1
                                    )
                                    for sp in range(bulk_wq_parameter_num):
                                        mass_incoming_mat[sp][0] = (
                                            mass_incoming_mat[sp][0]
                                            + abs(link_flow_rate)
                                            * link_conc_array_a[sp][pos][incoming_link]
                                        )
                                else:
                                    tank_outflow = tank_outflow + abs(link_flow_rate)
                        for o in range(num_outgoing_links):
                            outgoing_link = links_connecting_from_node[o]
                            if index_omitted_links.count(outgoing_link + 1) > 0:
                                break
                            else:
                                link_flow_rate = link_flow_rate_time_step[outgoing_link]
                                if link_flow_rate < 0:
                                    pos = 0
                                    for sp in range(bulk_wq_parameter_num):
                                        mass_incoming_mat[sp][0] = (
                                            mass_incoming_mat[sp][0]
                                            + abs(link_flow_rate)
                                            * link_conc_array_a[sp][pos][outgoing_link]
                                        )
                                else:
                                    tank_outflow = tank_outflow + abs(link_flow_rate)
                        for sp in range(bulk_wq_parameter_num):
                            if wq_step == 1:
                                tank_flow_volume_prev = tank_flow_volume
                                tank_initial_conc = node_conc_initial_mat[n][sp]
                            else:
                                tank_initial_conc = node_conc_array_a[sp][wq_step - 2][
                                    n
                                ]
                            # CSTR model
                            node_conc_array_a[sp][wq_step - 1][n] = (
                                tank_flow_volume_prev * tank_initial_conc
                                + mass_incoming_mat[sp][0] * wq_sim_time_step_s
                            ) / (tank_flow_volume + tank_outflow * wq_sim_time_step_s)
                        # Reactions
                        delta = module.tank_reaction(
                            wq_sim_time_step_s,
                            wq_step,
                            n,
                            tank_flow_volume_prev,
                            tank_flow_volume,
                            variable_values_mat[wq_iteration_count - 1],
                            node_conc_initial_mat,
                            node_conc_array_a,
                        )
                        for sp in range(bulk_wq_parameter_num):
                            node_conc_array_a[sp][wq_step - 1][n] = (
                                node_conc_array_a[sp][wq_step - 1][n] + delta[sp]
                            )

                        # Check for injection node
                        for sp in range(bulk_wq_parameter_num):
                            if (
                                index_injection_nodes.count(n + 1) > 0
                                and injection_quality_mat[
                                    index_injection_nodes.index(n + 1)
                                ][sp]
                                != 0
                            ):
                                if wq_step == 1:
                                    node_conc_array_a[sp][wq_step - 1][n] = (
                                        injection_quality_mat[
                                            index_injection_nodes.index(n + 1)
                                        ][sp]
                                    )
                                else:
                                    node_conc_array_a[sp][wq_step - 1][n] = (
                                        injection_quality_mat[
                                            index_injection_nodes.index(n + 1)
                                        ][sp]
                                        * injection_pattern[
                                            index_injection_nodes.index(n + 1)
                                        ][injection_pattern_step - 1][sp]
                                    )
                        node_conc_array_a[node_conc_array_a < 0] = 0
                    else:
                        node_outflow = abs(node_demand)
                        links_connecting_to_node = fn.incoming_links(n, end_node_mat)
                        links_connecting_from_node = fn.outgoing_links(
                            n, start_node_mat
                        )
                        num_incoming_links = len(links_connecting_to_node)
                        num_outgoing_links = len(links_connecting_from_node)
                        incoming_flow = 0
                        outgoing_flow = 0
                        mass_incoming_mat = np.zeros((bulk_wq_parameter_num, 1))
                        for i in range(num_incoming_links):
                            incoming_link = links_connecting_to_node[i]
                            if index_omitted_links.count(incoming_link + 1) > 0:
                                break
                            else:
                                link_flow_rate = link_flow_rate_time_step[incoming_link]
                                if link_flow_rate >= 0:
                                    pos = (
                                        int(
                                            link_conc_data_points_mat[wq_step - 1][
                                                incoming_link
                                            ]
                                        )
                                        - 1
                                    )
                                    for sp in range(bulk_wq_parameter_num):
                                        mass_incoming_mat[sp][0] = (
                                            mass_incoming_mat[sp][0]
                                            + abs(link_flow_rate)
                                            * link_conc_array_a[sp][pos][incoming_link]
                                        )
                                else:
                                    outgoing_flow = outgoing_flow + abs(link_flow_rate)
                        for o in range(num_outgoing_links):
                            outgoing_link = links_connecting_from_node[o]
                            if index_omitted_links.count(outgoing_link + 1) > 0:
                                break
                            else:
                                link_flow_rate = link_flow_rate_time_step[outgoing_link]
                                if link_flow_rate < 0:
                                    pos = (
                                        int(
                                            link_conc_data_points_mat[wq_step - 1][
                                                outgoing_link
                                            ]
                                        )
                                        - 1
                                    )
                                    for sp in range(bulk_wq_parameter_num):
                                        mass_incoming_mat[sp][0] = (
                                            mass_incoming_mat[sp][0]
                                            + abs(link_flow_rate)
                                            * link_conc_array_a[sp][pos][outgoing_link]
                                        )
                                else:
                                    outgoing_flow = outgoing_flow + abs(link_flow_rate)
                        total_outflow = node_outflow + outgoing_flow
                        if total_outflow == 0:
                            for sp in range(bulk_wq_parameter_num):
                                node_conc_array_a[sp][wq_step - 1][n] = 0
                        else:
                            for sp in range(bulk_wq_parameter_num):
                                node_conc_array_a[sp][wq_step - 1][n] = (
                                    mass_incoming_mat[sp][0] / total_outflow
                                )

                        # Check for injection node
                        for sp in range(bulk_wq_parameter_num):
                            if (
                                index_injection_nodes.count(n + 1) > 0
                                and injection_quality_mat[
                                    index_injection_nodes.index(n + 1)
                                ][sp]
                                != 0
                            ):
                                if wq_step == 1:
                                    node_conc_array_a[sp][wq_step - 1][n] = (
                                        injection_quality_mat[
                                            index_injection_nodes.index(n + 1)
                                        ][sp]
                                    )
                                else:
                                    node_conc_array_a[sp][wq_step - 1][n] = (
                                        injection_quality_mat[
                                            index_injection_nodes.index(n + 1)
                                        ][sp]
                                        * injection_pattern[
                                            index_injection_nodes.index(n + 1)
                                        ][injection_pattern_step - 1][sp]
                                    )
                    if flag_reservoir == 0:
                        for i in range(num_incoming_links):
                            incoming_link = links_connecting_to_node[i]
                            if index_omitted_links.count(incoming_link + 1) > 0:
                                break
                            else:
                                link_flow_rate = link_flow_rate_time_step[incoming_link]
                                if link_flow_rate >= 0:
                                    pos = (
                                        int(
                                            link_conc_data_points_mat[wq_step - 1][
                                                incoming_link
                                            ]
                                        )
                                        - 1
                                    )
                                else:
                                    pos = 0

                                for sp in range(bulk_wq_parameter_num):
                                    link_conc_array_a[sp][pos][incoming_link] = (
                                        node_conc_array_a[sp][wq_step - 1][n]
                                    )
                        for o in range(num_outgoing_links):
                            outgoing_link = links_connecting_from_node[o]
                            if index_omitted_links.count(outgoing_link + 1) > 0:
                                break
                            else:
                                link_flow_rate = link_flow_rate_time_step[outgoing_link]
                                if link_flow_rate >= 0:
                                    pos = 0
                                else:
                                    pos = (
                                        int(
                                            link_conc_data_points_mat[wq_step - 1][
                                                outgoing_link
                                            ]
                                        )
                                        - 1
                                    )

                                for sp in range(bulk_wq_parameter_num):
                                    link_conc_array_a[sp][pos][outgoing_link] = (
                                        node_conc_array_a[sp][wq_step - 1][n]
                                    )
            if count_nodes_lagrangian != (num_nodes - num_omitted_nodes):
                print("Error in Lagrangian stage dedicated to nodes!")
                exit()
            link_conc_array = copy.deepcopy(link_conc_array_a)
            node_conc_array = copy.deepcopy(node_conc_array_a)
            wq_time += wq_sim_time_step_s
        print(
            "Water quality simulation (Iteration %d) completed." % (wq_iteration_count)
        )
        # Creating folder to save simulation results
        folder_name = "Results_Iteration" + str(wq_iteration_count)
        path = os.path.join(getcwd(), folder_name)
        if os.path.exists(path):
            print(
                "Output folder already exists in the directory. No new folder created."
            )
        else:
            os.makedirs(path)
        # Reporting the water quality simulation results
        reporting_time_cycle_s = wq_sim_days * 24 * 3600
        reporting_time_steps = (reporting_time_cycle_s / wq_sim_time_step_s) + 1
        reporting_step_start = total_wq_steps - reporting_time_steps
        reporting_step_end = total_wq_steps
        node_conc_report = np.zeros(
            (bulk_wq_parameter_num, int(reporting_time_steps), num_nodes)
        )
        for sp in range(bulk_wq_parameter_num):
            node_conc_report[sp] = node_conc_array[sp][
                int(reporting_step_start) : int(reporting_step_end) + 1
            ]
        time_mat = np.zeros((int(reporting_time_steps), 1))
        for t in range(int(reporting_time_steps)):
            time_mat[t] = (t * wq_sim_time_step_s) / 3600
        data_time = pd.DataFrame(time_mat)
        # Printing the time versus node concentration as Excel
        file_path1 = os.path.join(path, "Time versus node_concentration.xlsx")
        w1 = pd.ExcelWriter(file_path1, engine="xlsxwriter")
        for sp in range(bulk_wq_parameter_num):
            data_conc = pd.DataFrame(node_conc_report[sp])
            data = pd.concat([data_time, data_conc], axis=1)
            data.to_excel(w1, sheet_name="Bulk parameter " + str(sp + 1))
        data_water_age = pd.DataFrame(
            Q.NodeQuality[int(reporting_step_start) : int(reporting_step_end) + 1]
        )
        data_water_age_out = pd.concat([data_time, data_water_age], axis=1)
        data_water_age_out.to_excel(w1, sheet_name="Water age")
        w1.close()
        print("Excel files created.")
        wq_iteration_count += 1
    file_path2 = "Quality Input values.xlsx"
    w2 = pd.ExcelWriter(file_path2, engine="xlsxwriter")
    for i in range(num_reservoirs):
        data_source_quality = pd.DataFrame(reservoir_quality[i])
        data_source_quality.to_excel(w2, sheet_name="Source Quality " + str(i + 1))
    if len(injection_quality) != 0:
        for i in range(len(injection_quality)):
            data_injection_quality = pd.DataFrame(injection_quality[i])
            data_injection_quality.to_excel(
                w2, sheet_name="Injection Quality " + str(i + 1)
            )
    data_para = pd.DataFrame(variable_values_mat)
    data_para.to_excel(w2, sheet_name="Parameter values")
    w2.close()
    print("Anlysis completed...")
    print("Simulation time in seconds is %f" % (time.time() - start_time))
    exit()



from setuptools import setup, find_packages
import pathlib
import subprocess

root = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (root / "README.md").read_text(encoding="utf-8")

setup(
    name="epytc",
    version="1.0.0",
    description="An open-source Python package for modeling water quality in water distribution systems",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SWIL-IITK/EPyT-C",
    author="Gopinbathan R Abhijith",
    author_email="abhijith@iitk.ac.in",
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python",
        "Operating System :: OS Independent",
    ],
    keywords="epanet, water, networks, quality, simulations, water quality modelling",
    python_requires=">=3.8",
    packages=find_packages(include=["epytc", "epytc.*"]),
    # include the defaul_values.yaml file
    package_data={"": ["*.yaml"]},
    install_requires=["epyt", "hydra-core"],
)
